from types import FunctionType, LambdaType, MethodType
from typing import Optional, Union

from whylogs.api.logger import log
from whylogs.core import DatasetSchema

FuncType = Union[FunctionType, MethodType, LambdaType]


def profiling(*, schema: Optional[DatasetSchema] = None):  # type: ignore
    def decorated(func: FuncType) -> FunctionType:
        def wrapper(*args, **kwargs):  # type: ignore
            df = func(*args, **kwargs)
            df.profiling_results = log(pandas=df, schema=schema)
            return df

        return wrapper  # type: ignore

    return decorated



from abc import ABC, abstractmethod
from typing import Dict, List, TypeVar

from typing_extensions import TypeAlias

from whylogs.core.datatypes import DataType, Fractional, Integral, String
from whylogs.core.metrics import StandardMetric
from whylogs.core.metrics.metrics import Metric

M = TypeVar("M", bound=Metric)
ColumnSchema: TypeAlias = "ColumnSchema"  # type: ignore


class Resolver(ABC):
    """A resolver maps from a column name and a data type to trackers.

    Note that the key of the result dictionaries defines the namespaces of the metrics in the serialized form."""

    @abstractmethod
    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        raise NotImplementedError


class StandardResolver(Resolver):
    """Standard metric resolution with builtin types."""

    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:

        metrics: List[StandardMetric] = [StandardMetric.counts, StandardMetric.types]

        if isinstance(why_type, Integral):
            metrics.append(StandardMetric.distribution)
            metrics.append(StandardMetric.ints)
            metrics.append(StandardMetric.cardinality)
            metrics.append(StandardMetric.frequent_items)
        elif isinstance(why_type, Fractional):
            metrics.append(StandardMetric.cardinality)
            metrics.append(StandardMetric.distribution)
        elif isinstance(why_type, String):  # Catch all category as we map 'object' here
            metrics.append(StandardMetric.cardinality)
            if column_schema.cfg.track_unicode_ranges:
                metrics.append(StandardMetric.unicode_range)
            metrics.append(StandardMetric.distribution)  # 'object' columns can contain Decimal
            metrics.append(StandardMetric.frequent_items)

        if column_schema.cfg.fi_disabled:
            metrics.remove(StandardMetric.frequent_items)

        result: Dict[str, Metric] = {}
        for m in metrics:
            result[m.name] = m.zero(column_schema.cfg)
        return result


class LimitedTrackingResolver(Resolver):
    """Resolver that skips frequent item and cardinality trackers."""

    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        metrics: List[StandardMetric] = [StandardMetric.counts, StandardMetric.types]
        if isinstance(why_type, Integral):
            metrics.append(StandardMetric.distribution)
            metrics.append(StandardMetric.ints)

        elif isinstance(why_type, Fractional):
            metrics.append(StandardMetric.distribution)

        result: Dict[str, Metric] = {}
        for m in metrics:
            result[m.name] = m.zero(column_schema.cfg)
        return result


class HistogramCountingTrackingResolver(Resolver):
    """Resolver that only adds distribution tracker."""

    def resolve(self, name: str, why_type: DataType, column_schema: ColumnSchema) -> Dict[str, Metric]:
        metrics: List[StandardMetric] = [StandardMetric.distribution]
        result: Dict[str, Metric] = {}
        for m in metrics:
            result[m.name] = m.zero(column_schema.cfg)
        return result



from typing import Any, Callable, Dict, Generic, Optional, TypeVar

import whylogs_sketching as ds  # type: ignore

from whylogs.core.metrics.decorators import (
    DecoratedFunction,
    FuncType,
    _decorate_func,
    _func_wrapper,
)
from whylogs.core.proto import (
    FrequentItemsSketchMessage,
    HllSketchMessage,
    KllSketchMessage,
    MetricComponentMessage,
)

M = TypeVar("M")
NUM = TypeVar("NUM", float, int)

_MAX_BUILT_IN_ID = 100


class Serializer(DecoratedFunction, Generic[M]):
    def __init__(self, *, func: FuncType, name: str):
        self._name = name
        self._func = func

    @property
    def name(self) -> str:
        return self._name

    def __call__(self, *, value: M) -> MetricComponentMessage:
        return self._func(value)

    @classmethod
    def build(cls, func: FuncType, name: str) -> "Serializer":  # noqa
        return Serializer(func=func, name=name)  # type: ignore


_TYPED_SERIALIZERS: Dict[type, Serializer] = {}
_INDEXED_SERIALIZERS: Dict[int, Serializer] = {}


def _builtin_serializer(*, name: str) -> Callable[[Callable], Serializer]:
    """Decorator for a builtin field aggregator.

    Note that since these are built in, they MUST be unique and stable over time.
    Args:
        name: a human readable string. This must be unique for validation.

    """

    def decorated(func: FuncType) -> Serializer:
        annotations: Dict[str, type] = func.__annotations__.copy()
        if annotations["return"] != MetricComponentMessage:
            raise ValueError("Invalid function type: return type is not MetricComponentMessage")
        annotations.pop("return")
        arg_len = len(annotations.items())
        if arg_len != 1:
            raise ValueError(f"Expected 1 argument, got: {arg_len}")

        ser = Serializer[Any](func=func, name=name)
        input_type = next(iter(annotations.values()))

        _func_wrapper(
            func=func,
            key=input_type,
            name=f"builtin.{input_type}",
            wrapper_dict=_TYPED_SERIALIZERS,
            clazz=Serializer,
        )

        return ser

    return decorated  # type: ignore


@_builtin_serializer(name="n")
def _int(value: int) -> MetricComponentMessage:
    return MetricComponentMessage(n=int(value))


@_builtin_serializer(name="d")
def _float(value: float) -> MetricComponentMessage:
    return MetricComponentMessage(d=value)


@_builtin_serializer(name="kll")
def _kll(sketch: ds.kll_doubles_sketch) -> MetricComponentMessage:
    return MetricComponentMessage(kll=KllSketchMessage(sketch=sketch.serialize()))


@_builtin_serializer(name="kll")
def _hll_merge(sketch: ds.hll_sketch) -> MetricComponentMessage:
    return MetricComponentMessage(hll=HllSketchMessage(sketch=sketch.serialize_compact()))


@_builtin_serializer(name="kll")
def _fs_merge(sketch: ds.frequent_strings_sketch) -> MetricComponentMessage:
    msg = FrequentItemsSketchMessage(sketch=sketch.serialize())

    return MetricComponentMessage(frequent_items=msg)


class SerializerRegistry:
    def __init__(self) -> None:
        self._typed_serializers = _TYPED_SERIALIZERS.copy()
        self._id_serializer = _INDEXED_SERIALIZERS.copy()

    def get(self, *, mtype: Optional[type] = None, type_id: int = 0) -> Optional[Serializer]:
        if mtype is None and type_id <= 0:
            raise ValueError("Please specify mtype or id")
        res: Optional[Serializer] = None
        if type_id > 0:
            res = self._id_serializer.get(type_id)

        if res is None and mtype is not None:
            res = self._typed_serializers.get(mtype)

        return res


_STANDARD_REGISTRY = SerializerRegistry()


def get_serializer(
    *, mtype: Optional[type] = None, type_id: int = 0, registry: Optional[SerializerRegistry] = None
) -> Optional[Serializer]:
    if registry is None:
        registry = _STANDARD_REGISTRY
    return registry.get(mtype=mtype, type_id=type_id)


def serializer(*, type_id: int, registry: SerializerRegistry = _STANDARD_REGISTRY):  # type: ignore
    if type_id < _MAX_BUILT_IN_ID:
        raise ValueError("Custom aggregator id must be equal or greater than 100")

    if registry is None:
        registry = _STANDARD_REGISTRY

    return _decorate_func(key=type_id, name=f"custom.{type_id}", wrapper_dict=registry._id_serializer, clazz=Serializer)



import os
import re
from datetime import date, datetime, timedelta
from typing import Tuple, Union

import pandas as pd

from whylogs.datasets.configs import BaseConfig

base_config = BaseConfig()


def _adjust_date(current_date: str, date_offset: timedelta) -> datetime:
    adjusted_date = _validate_timestamp(current_date) + date_offset
    return adjusted_date


def _parse_interval(interval: str) -> Tuple[int, str]:
    try:
        result = re.findall(r"^(\d+)([DdMmHhSs])", interval)[0]
        return (int(result[0]), result[1].upper())
    except IndexError:
        raise ValueError("Could not parse interval {}!".format(interval))


def _adjust_df_date(df: pd.DataFrame, new_start_date: datetime) -> pd.DataFrame:
    df = _change_df_date_by_offset(df, new_start_date=new_start_date)
    df = df.set_index(["date"], drop=False)
    return df


def _change_df_date_by_offset(df: pd.DataFrame, new_start_date: datetime) -> pd.DataFrame:
    original_start_date = _validate_timestamp(df["date"][0])
    date_offset = new_start_date - original_start_date
    df["date"] = df["date"].apply(lambda x: _adjust_date(x, date_offset))
    return df


def _validate_timestamp(timestamp: Union[date, datetime, pd.Timestamp, str]) -> datetime:
    if isinstance(timestamp, pd.Timestamp):
        return timestamp.to_pydatetime()
    if isinstance(timestamp, str):
        try:
            return datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S%z")
        except ValueError:
            raise ValueError('Could not parse string as datetime. Expected format: "%Y-%m-%d %H:%M:%S%z"')
    if isinstance(timestamp, datetime):
        return timestamp
    if isinstance(timestamp, date):
        timestamp = datetime(
            year=timestamp.year,
            month=timestamp.month,
            day=timestamp.day,
        )
        return timestamp
    raise ValueError("You must pass either a Datetime or Date object to timestamp!")


def _get_data_home() -> str:
    home_path = os.path.join(".", base_config.data_folder)
    if not os.path.exists(home_path):
        os.makedirs(home_path)
    return home_path


def _get_dataset_path(folder_name: str) -> str:
    home_path = _get_data_home()
    dataset_path = os.path.join(home_path, folder_name)
    if not os.path.exists(dataset_path):
        os.makedirs(dataset_path)
    return dataset_path



import html
from abc import ABC, abstractmethod
from typing import Any, Optional

from IPython.core.display import HTML  # type: ignore

from whylogs import DatasetProfileView
from whylogs.api.writer import Writers
from whylogs.api.writer.writer import Writable, Writer
from whylogs.viz.enums.enums import PageSpec


class HTMLReport(Writable, ABC):
    def __init__(
        self,
        ref_view: Optional[DatasetProfileView] = None,
        target_view: Optional[DatasetProfileView] = None,
        height: Optional[str] = None,
    ):
        self.ref_view = ref_view
        self.target_view = target_view
        self.height = height or None

    def writer(self, name: str = "local") -> "HTMLReportWriter":
        writer = Writers.get(name)
        return HTMLReportWriter(report=self, writer=writer)

    def display(self, template: str, page_spec: PageSpec) -> HTML:
        if not self.height:
            self.height = page_spec.height
        iframe = f"""<div></div><iframe srcdoc="{html.escape(template)}" width=100% height={self.height}
        frameBorder=0></iframe>"""
        display = HTML(iframe)
        return display

    @abstractmethod
    def report(self) -> HTML:
        pass

    def write(self, path: Optional[str] = None, **kwargs: Any) -> None:
        """Create HTML file for a given report.

        Parameters
        ----------
        path: str, optional
            The path where the HTML reports will be stored to.

        Examples
        --------
        .. code-block:: python

            from whylogs.viz import VizProfile, SummaryDriftReport

            report = SummaryDriftReport.report()

            viz_profile = VizProfile(report=report)
            viz_profile.write(path="path/to/report/Report.html")
        """
        path = path or self.get_default_path()
        _html = self.report()
        _rendered_html = _html.data
        with self._safe_open_write(path) as file:
            file.write(_rendered_html)

    def option(self):
        return self

    def get_default_path(self) -> str:
        path = "html_reports/ProfileReport.html"
        return path


class HTMLReportWriter(object):
    def __init__(self, report: HTMLReport, writer: Writer) -> None:
        self._report = report
        self._writer = writer

    def option(self, **kwargs) -> "HTMLReportWriter":
        self._writer.option(**kwargs)
        return self

    def write(self, **kwargs: Any) -> None:
        self._writer.write(file=self._report, **kwargs)



import math
from logging import getLogger
from typing import List, Optional, Tuple, Union

import numpy as np
from typing_extensions import TypedDict

from whylogs.core.metrics import DistributionMetric
from whylogs.core.view.column_profile_view import ColumnProfileView

logger = getLogger(__name__)

# the kll_floats_sketch use 32bit floats, so we check precision against np.float32
FLOAT_MANTISSA_BITS = np.finfo(np.float32).nmant


class QuantileStats(TypedDict):
    iqr: float
    q1: float
    q3: float
    ninety_fifth_percentile: float
    fifth_percentile: float
    median: float


def _calculate_quantile_statistics(column_view: Union[ColumnProfileView, None]) -> Union[None, QuantileStats]:
    if column_view is None:
        return None

    distribution_metric: Optional[DistributionMetric] = column_view.get_metric("distribution")
    desired_quantiles = [0.05, 0.25, 0.5, 0.75, 0.95]

    if distribution_metric is None or distribution_metric.kll.value.is_empty():
        return None

    quantiles = distribution_metric.kll.value.get_quantiles(desired_quantiles)

    quantile_statistics: QuantileStats = {
        "fifth_percentile": quantiles[0],
        "iqr": quantiles[3] - quantiles[1],
        "q1": quantiles[1],
        "median": quantiles[2],
        "q3": quantiles[3],
        "ninety_fifth_percentile": quantiles[4],
    }
    return quantile_statistics


def _resize_bins(start: float, end: float, min_interval: float, width: float, n_buckets: int) -> Tuple[int, float]:
    new_buckets = math.floor((end - start) / min_interval)
    logger.warning(
        f"A bin width of {width} won't work with values in range of [{start}, {end}] "
        f"because numbers closer to each other than {int(min_interval)} might not be distinct "
        "when passed as float32: avoiding bin edge collisions by resizing from: "
        f"{n_buckets} to: {new_buckets} histogram buckets in summary."
    )
    n_buckets = max(new_buckets, 1)
    width = (end - start) / n_buckets
    logger.info(f"New bin width is: {width} across {n_buckets} buckets")
    return n_buckets, width


def _get_min_interval(abs_start: float, abs_end: float) -> float:
    """
    Figure out the floating point precision at the scale of the bin boundaries
    min_interval is the smallest difference between floats at this scale.
    """
    max_magnitude = max(abs_start, abs_end)
    log_min_interval = math.floor(math.log2(max_magnitude)) - FLOAT_MANTISSA_BITS
    min_interval = math.pow(2, log_min_interval)

    return min_interval


def _calculate_bins(
    end: float, start: float, n: int, avg_per_bucket: float, max_buckets: int
) -> Tuple[List[float], float]:
    # Include the max value in the right-most bin
    end += abs(end) * 1e-7
    abs_end = abs(end)
    abs_start = abs(start)

    # Include the right edge in the bin edges
    n_buckets = min(math.ceil(n / avg_per_bucket), max_buckets)
    width = (end - start) / n_buckets

    min_interval = _get_min_interval(abs_start, abs_end)

    # If the bin width is smaller than min_interval, we need bigger bins
    if width < min_interval:
        n_buckets, width = _resize_bins(start, end, min_interval, width, n_buckets)

    # Calculate histograms from the Probability Mass Function
    bins = [start + i * width for i in range(n_buckets + 1)]
    logger.debug(f"about to get pmf using start: {start} end:{end} width:{width} and n_buckets:{n_buckets}")
    logger.debug(f"bin: {bins}")
    return bins, end



from ..metric_constraints import MetricConstraint, MetricsSelector

# TODO implement skip_missing for all methods


def count_below_number(column_name: str, number: int) -> MetricConstraint:
    """Number of elements in a column must be below given number.

    Parameters
    ----------
    column_name : str
        Column the constraint is applied to
    number : float
        reference value for applying the constraint
    """

    def is_count_below(x) -> bool:
        return number >= x.n.value

    constraint = MetricConstraint(
        name=f"count of {column_name} lower than {number}",
        condition=is_count_below,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="counts"),
    )
    return constraint


def null_values_below_number(column_name: str, number: int) -> MetricConstraint:
    """Number of null values must be below given number.

    Parameters
    ----------
    column_name : str
        Column the constraint is applied to
    number : float
        reference value for applying the constraint
    """

    def is_null_below(x):
        return x.null.value < number

    constraint = MetricConstraint(
        name=f"null values of {column_name} lower than {number}",
        condition=is_null_below,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="counts"),
    )
    return constraint


def null_percentage_below_number(column_name: str, number: float) -> MetricConstraint:
    """Percentage of null values must be below given number.

    Parameters
    ----------
    column_name : str
        Column the constraint is applied to
    number : float
        reference value for applying the constraint
    """

    def is_null_percentage_below_number(x):
        return (x.null.value / x.n.value) <= number

    constraint = MetricConstraint(
        name=f"null percentage of {column_name} lower than {number}",
        condition=is_null_percentage_below_number,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="counts"),
    )
    return constraint



from typing import Any, Dict, Optional

from typing_extensions import Literal

from whylogs.api.logger.logger import Logger
from whylogs.api.logger.result_set import ProfileResultSet, ResultSet, ResultSetReader
from whylogs.api.logger.rolling import TimedRollingLogger
from whylogs.api.logger.transient import TransientLogger
from whylogs.core import DatasetProfile, DatasetSchema
from whylogs.core.model_performance_metrics.model_performance_metrics import (
    ModelPerformanceMetrics,
)
from whylogs.core.stubs import pd


def log(
    obj: Any = None,
    *,
    pandas: Optional[pd.DataFrame] = None,
    row: Optional[Dict[str, Any]] = None,
    schema: Optional[DatasetSchema] = None,
) -> ResultSet:
    return TransientLogger(schema=schema).log(obj, pandas=pandas, row=row)


def log_classification_metrics(
    data: pd.DataFrame,
    target_column: str,
    prediction_column: str,
    score_column: Optional[str] = None,
    schema: Optional[DatasetSchema] = None,
) -> ProfileResultSet:
    """
    Function to track metrics based on validation data.
    user may also pass the associated attribute names associated with
    target, prediction, and/or score.
    Parameters
    ----------
    targets : List[Union[str, bool, float, int]]
        actual validated values
    predictions : List[Union[str, bool, float, int]]
        inferred/predicted values
    scores : List[float], optional
        assocaited scores for each inferred, all values set to 1 if not
        passed
    """
    model_performance_metrics = ModelPerformanceMetrics()
    model_performance_metrics.compute_confusion_matrix(
        predictions=data[prediction_column].to_list(),
        targets=data[target_column].to_list(),
        scores=data[score_column].to_list() if score_column else None,
    )

    results = log(pandas=data, schema=schema)
    results.add_model_performance_metrics(model_performance_metrics)
    return results


def log_regression_metrics(
    data: pd.DataFrame,
    target_column: str,
    prediction_column: str,
    schema: Optional[DatasetSchema] = None,
) -> ProfileResultSet:
    """
    Function to track regression metrics based on validation data.
    user may also pass the associated attribute names associated with
    target, prediction, and/or score.
    Parameters
    ----------
    targets : List[Union[str, bool, float, int]]
        actual validated values
    predictions : List[Union[str, bool, float, int]]
        inferred/predicted values
    scores : List[float], optional
        assocaited scores for each inferred, all values set to 1 if not
        passed
    """
    model_performance_metrics = ModelPerformanceMetrics()
    model_performance_metrics.compute_regression_metrics(
        predictions=data[prediction_column].to_list(),
        targets=data[target_column].to_list(),
    )

    results = log(pandas=data, schema=schema)
    results.add_model_performance_metrics(model_performance_metrics)
    return results


def read(path: str) -> ResultSet:
    return ResultSet.read(path)


def reader(name: str) -> ResultSetReader:
    return ResultSet.reader(name=name)


def write(profile: DatasetProfile, base_dir: str) -> None:
    return profile.write(base_dir)


def logger(
    schema: Optional[DatasetSchema] = None,
    *,
    mode: Literal["transient", "rolling"] = "transient",
    **kwargs: Any,
) -> Logger:
    """

    Returns:
        Logger: a logger object
    """
    if mode == "transient":
        return TransientLogger(schema)
    elif mode == "rolling":
        return TimedRollingLogger(schema, **kwargs)
    else:
        raise ValueError("Unsupported logger mode")


__ALL__ = [
    "log",
    "read",
    "reader",
    "write",
    "logger",
    "Logger",
    "ResultSet",
    "RollingLogger",
    "SegmentedResultSet",
]



# from typing import Callable, Dict, Type, TypeVar
#
# UM_LAMBDA = Callable[[], UpdatableMetric]
# MM_LAMBDA = Callable[[], MergeableMetric]
#
# UM = TypeVar("UM", bound=UpdatableMetric)
# MM = TypeVar("MM", bound=MergeableMetric)
#
#
# class _MetricRegistry(object):
#     _updatable_metrics: Dict[str, UM_LAMBDA] = {}
#     _mergeable_metrics: Dict[str, MM_LAMBDA] = {}
#
#     def get_updatable_metric(self, m_name: str) -> UpdatableMetric:
#         return self._updatable_metrics[m_name]()
#
#     def get_mergeable_metric(self, m_name: str) -> MergeableMetric:
#         return self._mergeable_metrics[m_name]()
#
#     def register_mergeable(self, name: str, func: MM_LAMBDA) -> None:
#         if self._updatable_metrics.get(name) is not None:
#             raise ValueError(f"Mergeable metric already registered for {name}")
#         self._mergeable_metrics[name] = func
#
#     def register_updatable(self, name: str, func: UM_LAMBDA) -> None:
#         if self._updatable_metrics.get(name) is not None:
#             raise ValueError(f"Updatable metric already registered for {name}")
#         self._updatable_metrics[name] = func
#
#     def register(self, name: str, m_func: MM_LAMBDA, u_func: UM_LAMBDA) -> None:
#         self.register_mergeable(name, m_func)
#         self.register_updatable(name, u_func)
#
#     def register_types(self, name: str, u_type: Type[UM], m_type: Type[MM]) -> None:
#         self.register_mergeable(name, lambda: m_type())  # type: ignore
#         self.register_updatable(name, lambda: u_type())  # type: ignore
#
#
# _INSTANCE = _MetricRegistry()
#
#
# def get_registry() -> _MetricRegistry:
#     return _INSTANCE



from tempfile import NamedTemporaryFile
from typing import Optional

import boto3

from whylogs import ResultSet
from whylogs.api.reader.reader import Reader


class S3Reader(Reader):
    """
    A WhyLogs reader to fetch written profiles from an Amazon S3 bucket.

    >**IMPORTANT**: In order to correctly connect to your Amazon S3 bucket, make sure you have
    the following environment variables set: `[AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY]`.

    Parameters
    ----------
    bucket_name: str, optional
        The name of the bucket to connect to. Made optional so the user can also access it
        via the `option` method
    object_name: str, optional
        The s3's object name. It basically states the location where the file goes to.
        Also made optional, so it can be defined through the `option` method

    Examples
    --------
    An example usage of this method can be represented with the simple code below.

    ```python
    import whylogs as why

    profile = why.reader("s3").option(bucket_name="my_bucket", object_name="my/object.bin").read()
    ```
    """

    def __init__(self, object_name: Optional[str] = None, bucket_name: Optional[str] = None):
        self.s3_client = boto3.client("s3")
        self.object_name = object_name or None
        self.bucket_name = bucket_name or ""

    def read(self, **kwargs) -> ResultSet:
        with NamedTemporaryFile() as tmp_file:
            self.s3_client.download_file(Bucket=self.bucket_name, Key=self.object_name, Filename=tmp_file.name)
            tmp_file.flush()
            result_set = self.get_file_from_path(path=tmp_file.name)
        return result_set

    def option(self, object_name: Optional[str] = None, bucket_name: Optional[str] = None) -> None:
        if object_name is not None:
            self.object_name = object_name
        if bucket_name is not None:
            self.bucket_name = bucket_name



import math
from typing import List

from whylogs.core.proto.v0 import RegressionMetricsMessage


class RegressionMetrics:
    def __init__(self):
        self.count = 0
        self.sum_abs_diff = 0.0
        self.sum_diff = 0.0
        self.sum2_diff = 0.0

    def add(self, predictions: List[float], targets: List[float]):
        """
        Function adds predictions and targets computation of regression metrics.

        Args:
            predictions (List[float]):
            targets (List[float]):
        """

        # need to vectorize this
        for idx, target in enumerate(targets):

            self.sum_abs_diff += abs(predictions[idx] - target)
            self.sum_diff += predictions[idx] - target
            self.sum2_diff += (predictions[idx] - target) ** 2
            # To add later
            # self.nt_diff.track(predictions[idx] - target)
            self.count += 1

    def mean_absolute_error(self):
        if self.count == 0:
            return None
        return self.sum_abs_diff / self.count

    def mean_squared_error(self):
        if self.count == 0:
            return None
        return self.sum2_diff / self.count

    def root_mean_squared_error(self):
        if self.count == 0:
            return None
        return math.sqrt(self.sum2_diff / self.count)

    def merge(self, other):
        """
        Merge two seperate regression metrics.

        Args:
              other : regression metrics to merge with self
        Returns:
              RegressionMetrics: merged regression metrics
        """
        if other is None:
            return self

        if self.count == 0:
            return other
        if other.count == 0:
            return self

        new_reg = RegressionMetrics()
        new_reg.count = self.count + other.count
        new_reg.sum_abs_diff = self.sum_abs_diff + other.sum_abs_diff
        new_reg.sum_diff = self.sum_diff + other.sum_diff
        new_reg.sum2_diff = self.sum2_diff + other.sum2_diff

        return new_reg

    def to_protobuf(
        self,
    ):
        """
        Convert to protobuf

        Returns:
            TYPE: Protobuf Message
        """

        return RegressionMetricsMessage(
            prediction_field="0",
            target_field="0",
            count=self.count,
            sum_abs_diff=self.sum_abs_diff,
            sum_diff=self.sum_diff,
            sum2_diff=self.sum2_diff,
        )

    @classmethod
    def from_protobuf(
        cls,
        message: RegressionMetricsMessage,
    ):
        if message.ByteSize() == 0:
            return None

        reg_met = RegressionMetrics()
        reg_met.count = message.count
        reg_met.sum_abs_diff = message.sum_abs_diff
        reg_met.sum_diff = message.sum_diff
        reg_met.sum2_diff = message.sum2_diff

        return reg_met



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v0_constraints.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import v0_messages_pb2 as v0__messages__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2
from . import v0_summaries_pb2 as v0__summaries__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14v0_constraints.proto\x1a\x11v0_messages.proto\x1a\x1cgoogle/protobuf/struct.proto\x1a\x12v0_summaries.proto\"\xcf\x03\n\x14SummaryConstraintMsg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x13\n\x0b\x66irst_field\x18\x02 \x01(\t\x12\x16\n\x0csecond_field\x18\x03 \x01(\tH\x00\x12\x0f\n\x05value\x18\x04 \x01(\x01H\x00\x12/\n\x07\x62\x65tween\x18\x07 \x01(\x0b\x32\x1c.SummaryBetweenConstraintMsgH\x00\x12\x33\n\rreference_set\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.ListValueH\x00\x12\x13\n\tvalue_str\x18\x0c \x01(\tH\x00\x12\x0f\n\x02op\x18\x05 \x01(\x0e\x32\x03.Op\x12\x0f\n\x07verbose\x18\x06 \x01(\x08\x12\x16\n\x0equantile_value\x18\t \x01(\x01\x12J\n\x17\x63ontinuous_distribution\x18\n \x01(\x0b\x32\'.ReferenceDistributionContinuousMessageH\x01\x12\x46\n\x15\x64iscrete_distribution\x18\x0b \x01(\x0b\x32%.ReferenceDistributionDiscreteMessageH\x01\x42\x08\n\x06secondB\x18\n\x16reference_distribution\"Q\n&ReferenceDistributionContinuousMessage\x12\'\n\x06sketch\x18\x01 \x01(\x0b\x32\x17.KllFloatsSketchMessage\"\x95\x01\n$ReferenceDistributionDiscreteMessage\x12-\n\x0e\x66requent_items\x18\x01 \x01(\x0b\x32\x15.FrequentItemsSummary\x12)\n\x0cunique_count\x18\x02 \x01(\x0b\x32\x13.UniqueCountSummary\x12\x13\n\x0btotal_count\x18\x03 \x01(\x02\"\x8c\x01\n\x1bSummaryBetweenConstraintMsg\x12\x16\n\x0csecond_field\x18\x01 \x01(\tH\x00\x12\x15\n\x0blower_value\x18\x02 \x01(\x01H\x00\x12\x15\n\x0bthird_field\x18\x03 \x01(\tH\x01\x12\x15\n\x0bupper_value\x18\x04 \x01(\x01H\x01\x42\x07\n\x05lowerB\x07\n\x05upper\"=\n\x10\x41pplyFunctionMsg\x12\x10\n\x08\x66unction\x18\x01 \x01(\t\x12\x17\n\x0freference_value\x18\x02 \x01(\t\"\xf7\x01\n\x12ValueConstraintMsg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x0f\n\x05value\x18\x02 \x01(\x01H\x00\x12\x17\n\rregex_pattern\x18\x05 \x01(\tH\x00\x12/\n\tvalue_set\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.ListValueH\x00\x12%\n\x08\x66unction\x18\x07 \x01(\x0b\x32\x11.ApplyFunctionMsgH\x00\x12\x0f\n\x02op\x18\x03 \x01(\x0e\x32\x03.Op\x12\x0f\n\x07verbose\x18\x04 \x01(\x08\x12\r\n\x05total\x18\x08 \x01(\r\x12\x10\n\x08\x66\x61ilures\x18\t \x01(\rB\x0e\n\x0csecond_field\"\x86\x03\n\x1dMultiColumnValueConstraintMsg\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x37\n\x11\x64\x65pendent_columns\x18\x02 \x01(\x0b\x32\x1a.google.protobuf.ListValueH\x00\x12\x1a\n\x10\x64\x65pendent_column\x18\x03 \x01(\tH\x00\x12\x0f\n\x05value\x18\x06 \x01(\x01H\x01\x12/\n\tvalue_set\x18\x07 \x01(\x0b\x32\x1a.google.protobuf.ListValueH\x01\x12\x37\n\x11reference_columns\x18\x08 \x01(\x0b\x32\x1a.google.protobuf.ListValueH\x01\x12\x0f\n\x02op\x18\x04 \x01(\x0e\x32\x03.Op\x12\x0f\n\x07verbose\x18\x05 \x01(\x08\x12*\n\x1dinternal_dependent_columns_op\x18\t \x01(\x0e\x32\x03.Op\x12\r\n\x05total\x18\n \x01(\r\x12\x10\n\x08\x66\x61ilures\x18\x0b \x01(\rB\x0b\n\tdependentB\x0b\n\treference\"\x81\x01\n\x13ValueConstraintMsgs\x12(\n\x0b\x63onstraints\x18\x01 \x03(\x0b\x32\x13.ValueConstraintMsg\x12@\n\x18multi_column_constraints\x18\x02 \x03(\x0b\x32\x1e.MultiColumnValueConstraintMsg\"C\n\x15SummaryConstraintMsgs\x12*\n\x0b\x63onstraints\x18\x01 \x03(\x0b\x32\x15.SummaryConstraintMsg\"\xed\x03\n\x14\x44\x61tasetConstraintMsg\x12(\n\nproperties\x18\x01 \x01(\x0b\x32\x14.DatasetPropertiesV0\x12\x46\n\x11value_constraints\x18\x02 \x03(\x0b\x32+.DatasetConstraintMsg.ValueConstraintsEntry\x12J\n\x13summary_constraints\x18\x03 \x03(\x0b\x32-.DatasetConstraintMsg.SummaryConstraintsEntry\x12\x37\n\x17table_shape_constraints\x18\x04 \x01(\x0b\x32\x16.SummaryConstraintMsgs\x12<\n\x1emulti_column_value_constraints\x18\x05 \x01(\x0b\x32\x14.ValueConstraintMsgs\x1aM\n\x15ValueConstraintsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12#\n\x05value\x18\x02 \x01(\x0b\x32\x14.ValueConstraintMsgs:\x02\x38\x01\x1aQ\n\x17SummaryConstraintsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12%\n\x05value\x18\x02 \x01(\x0b\x32\x16.SummaryConstraintMsgs:\x02\x38\x01*\xc5\x01\n\x02Op\x12\n\n\x06unused\x10\x00\x12\x06\n\x02LT\x10\x01\x12\x06\n\x02LE\x10\x02\x12\x06\n\x02\x45Q\x10\x03\x12\x06\n\x02NE\x10\x04\x12\x06\n\x02GE\x10\x05\x12\x06\n\x02GT\x10\x06\x12\t\n\x05MATCH\x10\x07\x12\x0b\n\x07NOMATCH\x10\x08\x12\x08\n\x04\x42TWN\x10\t\x12\n\n\x06IN_SET\x10\n\x12\x0f\n\x0b\x43ONTAIN_SET\x10\x0b\x12\n\n\x06\x45Q_SET\x10\x0c\x12\x0e\n\nAPPLY_FUNC\x10\r\x12\x06\n\x02IN\x10\x0e\x12\x0b\n\x07\x43ONTAIN\x10\x0f\x12\n\n\x06NOT_IN\x10\x10\x12\x07\n\x03SUM\x10\x11\x42/\n\x1e\x63om.whylogs.v0.core.constraintB\x0b\x43onstraintsP\x01\x62\x06proto3')

_OP = DESCRIPTOR.enum_types_by_name['Op']
Op = enum_type_wrapper.EnumTypeWrapper(_OP)
unused = 0
LT = 1
LE = 2
EQ = 3
NE = 4
GE = 5
GT = 6
MATCH = 7
NOMATCH = 8
BTWN = 9
IN_SET = 10
CONTAIN_SET = 11
EQ_SET = 12
APPLY_FUNC = 13
IN = 14
CONTAIN = 15
NOT_IN = 16
SUM = 17


_SUMMARYCONSTRAINTMSG = DESCRIPTOR.message_types_by_name['SummaryConstraintMsg']
_REFERENCEDISTRIBUTIONCONTINUOUSMESSAGE = DESCRIPTOR.message_types_by_name['ReferenceDistributionContinuousMessage']
_REFERENCEDISTRIBUTIONDISCRETEMESSAGE = DESCRIPTOR.message_types_by_name['ReferenceDistributionDiscreteMessage']
_SUMMARYBETWEENCONSTRAINTMSG = DESCRIPTOR.message_types_by_name['SummaryBetweenConstraintMsg']
_APPLYFUNCTIONMSG = DESCRIPTOR.message_types_by_name['ApplyFunctionMsg']
_VALUECONSTRAINTMSG = DESCRIPTOR.message_types_by_name['ValueConstraintMsg']
_MULTICOLUMNVALUECONSTRAINTMSG = DESCRIPTOR.message_types_by_name['MultiColumnValueConstraintMsg']
_VALUECONSTRAINTMSGS = DESCRIPTOR.message_types_by_name['ValueConstraintMsgs']
_SUMMARYCONSTRAINTMSGS = DESCRIPTOR.message_types_by_name['SummaryConstraintMsgs']
_DATASETCONSTRAINTMSG = DESCRIPTOR.message_types_by_name['DatasetConstraintMsg']
_DATASETCONSTRAINTMSG_VALUECONSTRAINTSENTRY = _DATASETCONSTRAINTMSG.nested_types_by_name['ValueConstraintsEntry']
_DATASETCONSTRAINTMSG_SUMMARYCONSTRAINTSENTRY = _DATASETCONSTRAINTMSG.nested_types_by_name['SummaryConstraintsEntry']
SummaryConstraintMsg = _reflection.GeneratedProtocolMessageType('SummaryConstraintMsg', (_message.Message,), {
  'DESCRIPTOR' : _SUMMARYCONSTRAINTMSG,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:SummaryConstraintMsg)
  })
_sym_db.RegisterMessage(SummaryConstraintMsg)

ReferenceDistributionContinuousMessage = _reflection.GeneratedProtocolMessageType('ReferenceDistributionContinuousMessage', (_message.Message,), {
  'DESCRIPTOR' : _REFERENCEDISTRIBUTIONCONTINUOUSMESSAGE,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:ReferenceDistributionContinuousMessage)
  })
_sym_db.RegisterMessage(ReferenceDistributionContinuousMessage)

ReferenceDistributionDiscreteMessage = _reflection.GeneratedProtocolMessageType('ReferenceDistributionDiscreteMessage', (_message.Message,), {
  'DESCRIPTOR' : _REFERENCEDISTRIBUTIONDISCRETEMESSAGE,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:ReferenceDistributionDiscreteMessage)
  })
_sym_db.RegisterMessage(ReferenceDistributionDiscreteMessage)

SummaryBetweenConstraintMsg = _reflection.GeneratedProtocolMessageType('SummaryBetweenConstraintMsg', (_message.Message,), {
  'DESCRIPTOR' : _SUMMARYBETWEENCONSTRAINTMSG,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:SummaryBetweenConstraintMsg)
  })
_sym_db.RegisterMessage(SummaryBetweenConstraintMsg)

ApplyFunctionMsg = _reflection.GeneratedProtocolMessageType('ApplyFunctionMsg', (_message.Message,), {
  'DESCRIPTOR' : _APPLYFUNCTIONMSG,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:ApplyFunctionMsg)
  })
_sym_db.RegisterMessage(ApplyFunctionMsg)

ValueConstraintMsg = _reflection.GeneratedProtocolMessageType('ValueConstraintMsg', (_message.Message,), {
  'DESCRIPTOR' : _VALUECONSTRAINTMSG,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:ValueConstraintMsg)
  })
_sym_db.RegisterMessage(ValueConstraintMsg)

MultiColumnValueConstraintMsg = _reflection.GeneratedProtocolMessageType('MultiColumnValueConstraintMsg', (_message.Message,), {
  'DESCRIPTOR' : _MULTICOLUMNVALUECONSTRAINTMSG,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:MultiColumnValueConstraintMsg)
  })
_sym_db.RegisterMessage(MultiColumnValueConstraintMsg)

ValueConstraintMsgs = _reflection.GeneratedProtocolMessageType('ValueConstraintMsgs', (_message.Message,), {
  'DESCRIPTOR' : _VALUECONSTRAINTMSGS,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:ValueConstraintMsgs)
  })
_sym_db.RegisterMessage(ValueConstraintMsgs)

SummaryConstraintMsgs = _reflection.GeneratedProtocolMessageType('SummaryConstraintMsgs', (_message.Message,), {
  'DESCRIPTOR' : _SUMMARYCONSTRAINTMSGS,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:SummaryConstraintMsgs)
  })
_sym_db.RegisterMessage(SummaryConstraintMsgs)

DatasetConstraintMsg = _reflection.GeneratedProtocolMessageType('DatasetConstraintMsg', (_message.Message,), {

  'ValueConstraintsEntry' : _reflection.GeneratedProtocolMessageType('ValueConstraintsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETCONSTRAINTMSG_VALUECONSTRAINTSENTRY,
    '__module__' : 'v0_constraints_pb2'
    # @@protoc_insertion_point(class_scope:DatasetConstraintMsg.ValueConstraintsEntry)
    })
  ,

  'SummaryConstraintsEntry' : _reflection.GeneratedProtocolMessageType('SummaryConstraintsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETCONSTRAINTMSG_SUMMARYCONSTRAINTSENTRY,
    '__module__' : 'v0_constraints_pb2'
    # @@protoc_insertion_point(class_scope:DatasetConstraintMsg.SummaryConstraintsEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETCONSTRAINTMSG,
  '__module__' : 'v0_constraints_pb2'
  # @@protoc_insertion_point(class_scope:DatasetConstraintMsg)
  })
_sym_db.RegisterMessage(DatasetConstraintMsg)
_sym_db.RegisterMessage(DatasetConstraintMsg.ValueConstraintsEntry)
_sym_db.RegisterMessage(DatasetConstraintMsg.SummaryConstraintsEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\036com.whylogs.v0.core.constraintB\013ConstraintsP\001'
  _DATASETCONSTRAINTMSG_VALUECONSTRAINTSENTRY._options = None
  _DATASETCONSTRAINTMSG_VALUECONSTRAINTSENTRY._serialized_options = b'8\001'
  _DATASETCONSTRAINTMSG_SUMMARYCONSTRAINTSENTRY._options = None
  _DATASETCONSTRAINTMSG_SUMMARYCONSTRAINTSENTRY._serialized_options = b'8\001'
  _OP._serialized_start=2341
  _OP._serialized_end=2538
  _SUMMARYCONSTRAINTMSG._serialized_start=94
  _SUMMARYCONSTRAINTMSG._serialized_end=557
  _REFERENCEDISTRIBUTIONCONTINUOUSMESSAGE._serialized_start=559
  _REFERENCEDISTRIBUTIONCONTINUOUSMESSAGE._serialized_end=640
  _REFERENCEDISTRIBUTIONDISCRETEMESSAGE._serialized_start=643
  _REFERENCEDISTRIBUTIONDISCRETEMESSAGE._serialized_end=792
  _SUMMARYBETWEENCONSTRAINTMSG._serialized_start=795
  _SUMMARYBETWEENCONSTRAINTMSG._serialized_end=935
  _APPLYFUNCTIONMSG._serialized_start=937
  _APPLYFUNCTIONMSG._serialized_end=998
  _VALUECONSTRAINTMSG._serialized_start=1001
  _VALUECONSTRAINTMSG._serialized_end=1248
  _MULTICOLUMNVALUECONSTRAINTMSG._serialized_start=1251
  _MULTICOLUMNVALUECONSTRAINTMSG._serialized_end=1641
  _VALUECONSTRAINTMSGS._serialized_start=1644
  _VALUECONSTRAINTMSGS._serialized_end=1773
  _SUMMARYCONSTRAINTMSGS._serialized_start=1775
  _SUMMARYCONSTRAINTMSGS._serialized_end=1842
  _DATASETCONSTRAINTMSG._serialized_start=1845
  _DATASETCONSTRAINTMSG._serialized_end=2338
  _DATASETCONSTRAINTMSG_VALUECONSTRAINTSENTRY._serialized_start=2178
  _DATASETCONSTRAINTMSG_VALUECONSTRAINTSENTRY._serialized_end=2255
  _DATASETCONSTRAINTMSG_SUMMARYCONSTRAINTSENTRY._serialized_start=2257
  _DATASETCONSTRAINTMSG_SUMMARYCONSTRAINTSENTRY._serialized_end=2338
# @@protoc_insertion_point(module_scope)



from .column_profile_view import ColumnProfileView
from .dataset_profile_view import WHYLOGS_MAGIC_HEADER, DatasetProfileView

__ALL__ = [
    ColumnProfileView,
    DatasetProfileView,
    WHYLOGS_MAGIC_HEADER,
]



import IPython  # type: ignore  # noqa  # import ipython to make

from whylogs.viz.extensions.reports.summary_drift import SummaryDriftReport
from whylogs.viz.notebook_profile_viz import NotebookProfileVisualizer

__ALL__ = [
    # column
    NotebookProfileVisualizer,
    SummaryDriftReport,
]



from dataclasses import dataclass, field
from itertools import chain
from typing import Any, Callable, Dict, List

from whylogs.core.metrics.condition_count_metric import Condition
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.validators.validator import Validator


@dataclass
class ConditionValidator(Validator):
    conditions: Dict[str, Condition]
    actions: List[Callable[[str, str, Any], None]]
    name: str
    total: int = 0
    failures: Dict[str, int] = field(default_factory=dict)

    def __post_init__(self):
        for cond_name in self.conditions.keys():
            if cond_name not in self.failures:
                self.failures[cond_name] = 0

    def columnar_validate(self, data: PreprocessedColumn) -> None:
        count = 0
        count_failures = 0
        for x in list(chain.from_iterable(data.raw_iterator())):
            count += 1
            for cond_name, condition in self.conditions.items():
                if not condition.relation(x):
                    self.failures[cond_name] += 1
                    count_failures += 1
                    for action in self.actions:
                        action(self.name, cond_name, x)
        self.total = count

    def to_summary_dict(self) -> Dict[str, Any]:
        summary = {"total_evaluations": self.total}
        for cond_name, _ in self.conditions.items():
            summary[cond_name] = self.failures[cond_name]
        return summary



import dataclasses
import math
import statistics
import sys
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar, Union

import whylogs_sketching as ds  # type: ignore
from google.protobuf.struct_pb2 import Struct

from whylogs.core.configs import SummaryConfig
from whylogs.core.metrics.maths import (
    VarianceM2Result,
    parallel_variance_m2,
    welford_online_variance_m2,
)
from whylogs.core.metrics.metric_components import (
    FractionalComponent,
    FrequentStringsComponent,
    HllComponent,
    KllComponent,
    MaxIntegralComponent,
    MetricComponent,
    MinIntegralComponent,
)
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.proto import MetricComponentMessage, MetricMessage
from whylogs.core.stubs import pd as pd

T = TypeVar("T")
M = TypeVar("M", bound=MetricComponent)
NUM = TypeVar("NUM", float, int)

METRIC = TypeVar("METRIC", bound="Metric")


@dataclass(frozen=True)
class MetricConfig:
    hll_lg_k: int = 12
    kll_k: int = 256
    fi_lg_max_k: int = 10  # 128 entries
    fi_disabled: bool = False
    track_unicode_ranges: bool = False
    large_kll_k: bool = True
    kll_k_large: int = 1024
    unicode_ranges: Dict[str, Tuple[int, int]] = field(
        default_factory=lambda: {
            "emoticon": (0x1F600, 0x1F64F),
            "control": (0x00, 0x1F),
            "digits": (0x30, 0x39),
            "latin-lower": (0x41, 0x5A),
            "latin-upper": (0x61, 0x7A),
            "basic-latin": (0x00, 0x7F),
            "extended-latin": (0x0080, 0x02AF),
        }
    )
    lower_case: bool = True  # Convert Unicode characters to lower-case before counting Unicode ranges
    normalize: bool = True  # Unicode normalize strings before counting Unicode ranges


_METRIC_DESERIALIZER_REGISTRY: Dict[str, Type[METRIC]] = {}  # type: ignore


def custom_metric(metric: Type[METRIC], config: MetricConfig = MetricConfig()) -> Type[METRIC]:  # type: ignore
    global _METRIC_DESERIALIZER_REGISTRY
    _METRIC_DESERIALIZER_REGISTRY[metric.get_namespace(config)] = metric
    return metric


@dataclass(frozen=True)
class OperationResult:
    failures: int = 0
    successes: int = 0

    @classmethod
    def ok(cls, cnt: int = 1) -> "OperationResult":
        return OperationResult(successes=cnt, failures=0)

    @classmethod
    def failed(cls, cnt: int = 1) -> "OperationResult":
        return OperationResult(successes=0, failures=cnt)

    def __add__(self, other: "OperationResult") -> "OperationResult":
        return OperationResult(
            successes=self.successes + other.successes,
            failures=self.failures + other.failures,
        )


class Metric(ABC):
    @classmethod
    def get_namespace(cls, config: Optional[MetricConfig] = None) -> str:
        if config is None:
            config = MetricConfig()
        return cls.zero(config).namespace

    @property
    @abstractmethod
    def namespace(self) -> str:
        raise NotImplementedError

    def __post_init__(self):
        global _METRIC_DESERIALIZER_REGISTRY
        _METRIC_DESERIALIZER_REGISTRY[self.namespace] = self.__class__

    def __add__(self: METRIC, other: METRIC) -> METRIC:
        return self.merge(other)

    def merge(self: METRIC, other: METRIC) -> METRIC:
        res: Dict[str, MetricComponent] = {}
        for k, v in self.__dict__.items():
            if isinstance(v, MetricComponent):
                res[k] = v + other.__dict__[k]

        return self.__class__(**res)

    def to_protobuf(self) -> MetricMessage:
        if not dataclasses.is_dataclass(self):
            raise ValueError("Metric object is not a dataclass")

        res = {}
        for k, v in self.__dict__.items():
            if not isinstance(v, MetricComponent):
                continue
            res[k] = v.to_protobuf()
        return MetricMessage(metric_components=res)

    def get_component_paths(self) -> List[str]:
        res = []
        for k, v in self.__dict__.items():
            if not isinstance(v, MetricComponent):
                continue
            res.append(k)
        return res

    @abstractmethod
    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def zero(cls: Type[METRIC], config: MetricConfig) -> METRIC:
        pass

    @classmethod
    def from_protobuf(cls: Type[METRIC], msg: MetricMessage) -> METRIC:
        if not dataclasses.is_dataclass(cls):
            raise ValueError(f"Metric class: {cls} is not a dataclass")

        components: Dict[str, MetricComponent] = {}
        for k, m in msg.metric_components.items():
            components[k] = MetricComponent.from_protobuf(m)

        return cls(**components)


@dataclass(frozen=True)
class IntsMetric(Metric):
    max: MaxIntegralComponent
    min: MinIntegralComponent

    @property
    def namespace(self) -> str:
        return "ints"

    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        if data.len <= 0:
            return OperationResult.ok(0)

        successes = 0

        max_ = self.max.value
        min_ = self.min.value
        if data.numpy.ints is not None:
            max_ = max([max_, data.numpy.ints.max()])
            min_ = min([min_, data.numpy.ints.min()])
            successes += len(data.numpy.ints)

        if data.list.ints is not None:
            l_max = max(data.list.ints)
            l_min = min(data.list.ints)
            max_ = max([max_, l_max])
            min_ = min([min_, l_min])
            successes += len(data.list.ints)

        self.max.set(max_)
        self.min.set(min_)
        return OperationResult.ok(successes)

    @classmethod
    def zero(cls, config: MetricConfig) -> "IntsMetric":
        return IntsMetric(max=MaxIntegralComponent(-sys.maxsize), min=MinIntegralComponent(sys.maxsize))

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Union[int, float, str, None]]:
        return {"max": self.maximum, "min": self.minimum}

    @property
    def maximum(self) -> float:
        return self.max.value

    @property
    def minimum(self) -> float:
        return self.min.value


@dataclass(frozen=True)
class DistributionMetric(Metric):
    kll: KllComponent
    mean: FractionalComponent
    m2: FractionalComponent

    @property
    def namespace(self) -> str:
        return "distribution"

    def to_summary_dict(self, cfg: SummaryConfig = None) -> Dict[str, Union[int, float, str, None]]:
        if self.n == 0:
            quantiles = [None, None, None, None, None, None, None, None, None]
        else:
            quantiles = self.kll.value.get_quantiles([0.01, 0.05, 0.1, 0.25, 0.5, 0.75, 0.9, 0.95, 0.99])
        return {
            "mean": self.avg,
            "stddev": self.stddev,
            "n": self.kll.value.get_n(),
            "max": self.kll.value.get_max_value(),
            "min": self.kll.value.get_min_value(),
            "q_01": quantiles[0],
            "q_05": quantiles[1],
            "q_10": quantiles[2],
            "q_25": quantiles[3],
            "median": quantiles[4],
            "q_75": quantiles[5],
            "q_90": quantiles[6],
            "q_95": quantiles[7],
            "q_99": quantiles[8],
        }

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        """
        Update the operation

        Algorithm: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm

        Args:
            view: the preprocessed column

        Returns:
            how many successful operations we had
        """
        successes = 0

        first = VarianceM2Result(n=self.kll.value.get_n(), mean=self.mean.value, m2=self.m2.value)

        if view.numpy.len > 0:
            for arr in [view.numpy.floats, view.numpy.ints]:
                if arr is not None:
                    self.kll.value.update(arr)
                    n_b = len(arr)
                    if n_b > 1:
                        n_b = len(arr)
                        mean_b = arr.mean()
                        m2_b = arr.var() * (n_b - 1)

                        second = VarianceM2Result(n=n_b, mean=mean_b, m2=m2_b)
                        first = parallel_variance_m2(first=first, second=second)
                    elif n_b == 1:
                        # fall back to https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
                        # #Weighted_incremental_algorithm
                        if isinstance(arr, pd.Series):
                            first = welford_online_variance_m2(existing=first, new_value=arr.iloc[0])
                        else:
                            first = welford_online_variance_m2(existing=first, new_value=arr[0])

        for lst in [view.list.ints, view.list.floats]:
            if lst is not None and len(lst) > 0:
                self.kll.value.update_list(lst)
                n_b = len(lst)
                if n_b > 1:
                    mean_b = statistics.mean(lst)
                    m2_b = statistics.pvariance(lst) * (n_b - 1)
                    second = VarianceM2Result(n=n_b, mean=mean_b, m2=m2_b)

                    first = parallel_variance_m2(first=first, second=second)
                else:
                    first = welford_online_variance_m2(existing=first, new_value=lst[0])

        self.mean.set(first.mean)
        self.m2.set(first.m2)

        return OperationResult.ok(successes)

    def merge(self, other: "DistributionMetric") -> "DistributionMetric":
        # calculate mean and m2
        a_n = self.kll.value.get_n()
        b_n = other.kll.value.get_n()

        delta = other.avg - self.avg
        new_n = a_n + b_n
        if a_n != 0 or b_n != 0:
            m2 = self.m2.value + other.m2.value + delta**2 * a_n * b_n / new_n

            mean = (a_n / new_n) * (self.avg) + (b_n / new_n) * (other.avg)
        else:
            m2 = 0
            mean = 0

        # merge the sketch
        kll = self.kll + other.kll

        return DistributionMetric(kll=kll, mean=FractionalComponent(mean), m2=FractionalComponent(m2))

    @property
    def n(self) -> float:
        return self.kll.value.get_n()

    @property
    def variance(self) -> float:
        """Returns the population variance of the stream.

        https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
        """
        if self.n <= 1:
            return float(0)
        return self.m2.value / (self.kll.value.get_n() - 1)

    @property
    def stddev(self) -> float:
        if self.n <= 1:
            return float(0)
        return math.sqrt(self.variance)

    @property
    def avg(self) -> float:
        return self.mean.value

    @property
    def median(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.5])[0]
        return result

    @property
    def q_01(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.01])[0]
        return result

    @property
    def q_05(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.05])[0]
        return result

    @property
    def q_10(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.1])[0]
        return result

    @property
    def q_25(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.25])[0]
        return result

    @property
    def q_75(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.75])[0]
        return result

    @property
    def q_90(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.90])[0]
        return result

    @property
    def q_95(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.95])[0]
        return result

    @property
    def q_99(self) -> Optional[float]:
        result = None
        if self.n > 0:
            result = self.kll.value.get_quantiles([0.99])[0]
        return result

    @property
    def max(self) -> float:
        if self.kll.value.is_empty():
            return -sys.float_info.max
        return self.kll.value.get_max_value()

    @property
    def min(self) -> float:
        if self.kll.value.is_empty():
            return sys.float_info.max
        return self.kll.value.get_min_value()

    @classmethod
    def zero(cls, config: MetricConfig) -> "DistributionMetric":
        configured_kll_k = config.kll_k_large if config.large_kll_k else config.kll_k
        sk = ds.kll_doubles_sketch(k=configured_kll_k)
        return DistributionMetric(
            kll=KllComponent(sk),
            mean=FractionalComponent(0.0),
            m2=FractionalComponent(0.0),
        )


@dataclass(frozen=True)
class FrequentItem:
    value: str
    est: int
    upper: int
    lower: int


@dataclass(frozen=True)
class FrequentItemsMetric(Metric):
    frequent_strings: FrequentStringsComponent

    @property
    def namespace(self) -> str:
        return "frequent_items"

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        successes = 0
        for arr in [view.numpy.floats, view.numpy.ints]:
            if arr is not None:
                if isinstance(arr, pd.Series):
                    arr = arr.to_numpy()
                self.frequent_strings.value.update_np(arr)
                successes += len(arr)
        if view.pandas.strings is not None:
            self.frequent_strings.value.update_str_list(view.pandas.strings.to_list())
            successes += len(view.pandas.strings)

        if view.list.ints is not None:
            self.frequent_strings.value.update_int_list(view.list.ints)
            successes += len(view.list.ints)

        if view.list.floats is not None:
            self.frequent_strings.value.update_double_list(view.list.floats)
            successes += len(view.list.floats)

        if view.list.strings is not None:
            self.frequent_strings.value.update_str_list(view.list.strings)
            successes += len(view.list.strings)

        failures = 0
        if view.list.objs is not None:
            successes += len(view.list.objs)
        if view.pandas.objs is not None:
            successes += len(view.pandas.objs)

        return OperationResult(successes=successes, failures=failures)

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        all_freq_items = self.frequent_strings.value.get_frequent_items(
            cfg.frequent_items_error_type.to_datasketches_type()
        )
        if cfg.frequent_items_limit > 0:
            limited_freq_items = all_freq_items[: cfg.frequent_items_limit]
        else:
            limited_freq_items = all_freq_items
        items = [FrequentItem(value=x[0], est=x[1], upper=x[2], lower=x[3]) for x in limited_freq_items]
        return {"frequent_strings": items}

    @property
    def strings(self) -> List[FrequentItem]:
        if self.frequent_strings.value.is_empty():
            return []
        summary = self.to_summary_dict(cfg=SummaryConfig())
        return summary["frequent_strings"]

    @classmethod
    def zero(cls, config: MetricConfig) -> "FrequentItemsMetric":
        sk = ds.frequent_strings_sketch(lg_max_k=config.fi_lg_max_k)
        return FrequentItemsMetric(frequent_strings=FrequentStringsComponent(sk))


@dataclass(frozen=True)
class CardinalityMetric(Metric):
    hll: HllComponent

    @property
    def namespace(self) -> str:
        return "cardinality"

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        successes = 0
        if view.numpy.len > 0:
            if view.numpy.ints is not None:
                self.hll.value.update_np(view.numpy.ints)
                successes += len(view.numpy.ints)
            if view.numpy.floats is not None:
                self.hll.value.update_np(view.numpy.floats)
                successes += len(view.numpy.floats)
        if view.pandas.strings is not None:
            self.hll.value.update_str_list(view.pandas.strings.to_list())
            successes += len(view.pandas.strings)

        # update everything in the remaining lists
        if view.list.ints:
            self.hll.value.update_int_list(view.list.ints)
            successes += len(view.list.ints)
        if view.list.floats:
            self.hll.value.update_double_list(view.list.floats)
            successes += len(view.list.floats)
        if view.list.strings:
            self.hll.value.update_str_list(view.list.strings)
            successes += len(view.list.strings)

        failures = 0
        if view.list.objs:
            failures = len(view.list.objs)
        return OperationResult(successes=successes, failures=failures)

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        return {
            "est": self.hll.value.get_estimate(),
            f"upper_{cfg.hll_stddev}": self.hll.value.get_upper_bound(cfg.hll_stddev),
            f"lower_{cfg.hll_stddev}": self.hll.value.get_lower_bound(cfg.hll_stddev),
        }

    @property
    def estimate(self) -> Optional[float]:
        result = None
        if not self.hll.value.is_empty():
            result = self.hll.value.get_estimate()

        return result

    @property
    def upper_1(self) -> Optional[float]:
        result = None
        if not self.hll.value.is_empty():
            result = self.hll.value.get_upper_bound(1)

        return result

    @property
    def lower_1(self) -> Optional[float]:
        result = None
        if not self.hll.value.is_empty():
            result = self.hll.value.get_lower_bound(1)

        return result

    @classmethod
    def zero(cls, config: MetricConfig) -> "CardinalityMetric":
        sk = ds.hll_sketch(config.hll_lg_k)
        return CardinalityMetric(hll=HllComponent(sk))


def _drop_private_fields(data: List[Tuple[str, Any]]) -> Dict[str, Any]:
    return {k: v for k, v in data if not k.startswith("_")}


_STRUCT_NAME = "dataclass_param"


class CustomMetricBase(Metric, ABC):
    """
    You can use this as a base class for custom metrics that don't use
    the supplied or custom MetricComponents. Subclasses must be decorated with
    @dataclass. All fields not prefixed with an underscore will be included
    in the summary and will be [de]serialized. Such subclasses will need to
    implement the namespace, merge, and zero methods.
    """

    def get_component_paths(self) -> List[str]:
        return [_STRUCT_NAME]  # Assumes everything to be serde'd will be in the Struct

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        return asdict(self, dict_factory=_drop_private_fields)

    def to_protobuf(self) -> MetricMessage:
        mcm = MetricComponentMessage(dataclass_param=Struct())
        mcm.dataclass_param.update(asdict(self, dict_factory=_drop_private_fields))
        return MetricMessage(metric_components={_STRUCT_NAME: mcm})

    @classmethod
    def from_protobuf(cls: Type[METRIC], msg: MetricMessage) -> METRIC:
        my_struct = msg.metric_components[_STRUCT_NAME].dataclass_param
        return cls(**my_struct)



from dataclasses import dataclass
from typing import Generic, Optional, Type, TypeVar

import whylogs_sketching as ds  # type: ignore

from whylogs.core.metrics.aggregators import (
    AggregatorRegistry,
    _id_aggregator,
    get_aggregator,
)
from whylogs.core.metrics.deserializers import DeserializerRegistry, get_deserializer
from whylogs.core.metrics.serializers import SerializerRegistry, get_serializer
from whylogs.core.proto import MetricComponentMessage

T = TypeVar("T")
M = TypeVar("M", bound="MetricComponent")


@dataclass(frozen=True)
class Registries:
    aggregatorRegistry: Optional[AggregatorRegistry] = None
    serializerRegistry: Optional[SerializerRegistry] = None
    deserializerRegistry: Optional[DeserializerRegistry] = None


class MetricComponent(Generic[T]):
    """
    A metric component is the smallest unit for a metric.

    A metric might consist of multiple components. An example is distribution metric, which consists of kll sketch for
    histogram, mean and m2. The calculation of components could be independent or could be coupled with other
    components.
    """

    mtype: Optional[Type[T]]
    type_id: int = 0
    registries: Optional[Registries] = None

    def __init__(self, value: T):
        self._value = value

        registries = self.registries
        if registries is None:
            registries = Registries()

        self._registries = registries
        optional_mtype = None
        if hasattr(self, "mtype"):
            optional_mtype = self.mtype
        self._aggregator = get_aggregator(
            mtype=optional_mtype, type_id=self.type_id, registry=registries.aggregatorRegistry
        )
        self._serializer = get_serializer(
            mtype=optional_mtype, type_id=self.type_id, registry=registries.serializerRegistry
        )
        self._deserializer = get_deserializer(
            mtype=optional_mtype, type_id=self.type_id, registry=registries.deserializerRegistry
        )

        if self._serializer is None and self._deserializer is not None:
            raise ValueError("Serializer and deserializer must be defined in pairs, but serializer is None")
        if self._serializer is not None and self._deserializer is None:
            raise ValueError("Serializer and deserializer must be defined in pairs, but deserializer is None")

    @property
    def value(self) -> T:
        return self._value

    def set(self, value: T) -> None:
        self._value = value

    def __add__(self: M, other: M) -> M:
        if self._aggregator is None:
            raise ValueError(
                f"Attempting to aggregate metric component without an aggregator. Type: {self.mtype} with ID: "
                f"{self.type_id}"
            )
        return self.__class__(self._aggregator(self.value, other.value))

    def to_protobuf(self) -> MetricComponentMessage:
        if self._serializer is None:
            raise ValueError(
                f"Attempting to serialize metric component without a serializer. Type: {self.mtype} with ID: "
                f"{self.type_id}"
            )
        msg = self._serializer(value=self._value)
        msg.type_id = self.type_id
        return msg

    @classmethod
    def from_protobuf(cls: Type["M"], msg: MetricComponentMessage, registries: Optional[Registries] = None) -> "M":
        # TODO: need to deserialize this against a registry
        field = msg.WhichOneof("value")
        _mtype: Optional[type] = None
        if field == "n":
            _mtype = int
        elif field == "d":
            _mtype = float
        elif field == "frequent_items":
            _mtype = ds.frequent_strings_sketch
        elif field == "hll":
            _mtype = ds.hll_sketch
        elif field == "kll":
            _mtype = ds.kll_doubles_sketch
        _type_id = msg.type_id

        _registries = registries

        class DeserializedComponent(MetricComponent[M]):
            mtype = _mtype
            type_id = msg.type_id
            registries = _registries

        component = DeserializedComponent(value=None)  # type: ignore
        if component._deserializer is not None:
            component._value = component._deserializer(msg=msg)
            return component  # type: ignore
        else:
            raise ValueError(f"Attempt to deserialize a type without a deserializer. Type: {_mtype}. ID: {_type_id}")


class IntegralComponent(MetricComponent[int]):
    mtype = int


class MinIntegralComponent(MetricComponent[int]):
    mtype = int
    type_id = 1

    @staticmethod
    @_id_aggregator(type_id=type_id, name="min")
    def min(lhs: int, rhs: int) -> int:
        return min([lhs, rhs])


class MaxIntegralComponent(MetricComponent[int]):
    mtype = int
    type_id = 2

    @staticmethod
    @_id_aggregator(type_id=type_id, name="max")
    def max(lhs: int, rhs: int) -> int:
        return max([lhs, rhs])


class FractionalComponent(MetricComponent[float]):
    mtype = float


class KllComponent(MetricComponent[ds.kll_doubles_sketch]):
    mtype = ds.kll_doubles_sketch


class HllComponent(MetricComponent[ds.hll_sketch]):
    mtype = ds.hll_sketch


class FrequentStringsComponent(MetricComponent[ds.frequent_strings_sketch]):
    mtype = ds.frequent_strings_sketch


class CustomComponent(Generic[T], MetricComponent[T]):
    type_id = -1

    def __init__(self, value: T) -> None:
        if self.type_id < 100:
            raise ValueError("Custom metric component must have type index greater or equal to 100")
        super().__init__(value)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: v0_summaries.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import v0_messages_pb2 as v0__messages__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x12v0_summaries.proto\x1a\x11v0_messages.proto\x1a\x1cgoogle/protobuf/struct.proto\"D\n\x12UniqueCountSummary\x12\x10\n\x08\x65stimate\x18\x01 \x01(\x01\x12\r\n\x05upper\x18\x02 \x01(\x01\x12\r\n\x05lower\x18\x03 \x01(\x01\"~\n\x16\x46requentStringsSummary\x12\x33\n\x05items\x18\x01 \x03(\x0b\x32$.FrequentStringsSummary.FrequentItem\x1a/\n\x0c\x46requentItem\x12\r\n\x05value\x18\x01 \x01(\t\x12\x10\n\x08\x65stimate\x18\x02 \x01(\x01\"\x96\x02\n\x16\x46requentNumbersSummary\x12;\n\x07\x64oubles\x18\x01 \x03(\x0b\x32*.FrequentNumbersSummary.FrequentDoubleItem\x12\x37\n\x05longs\x18\x02 \x03(\x0b\x32(.FrequentNumbersSummary.FrequentLongItem\x1a\x43\n\x12\x46requentDoubleItem\x12\x10\n\x08\x65stimate\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x01\x12\x0c\n\x04rank\x18\x03 \x01(\x05\x1a\x41\n\x10\x46requentLongItem\x12\x10\n\x08\x65stimate\x18\x01 \x01(\x03\x12\r\n\x05value\x18\x02 \x01(\x03\x12\x0c\n\x04rank\x18\x03 \x01(\x05\"\x7f\n\x14\x46requentItemsSummary\x12\x31\n\x05items\x18\x01 \x03(\x0b\x32\".FrequentItemsSummary.FrequentItem\x1a\x34\n\x0c\x46requentItem\x12\x10\n\x08\x65stimate\x18\x01 \x01(\x03\x12\x12\n\njson_value\x18\x02 \x01(\t\"\xa2\x01\n\x0e\x43harPosSummary\x12\x16\n\x0e\x63haracter_list\x18\x01 \x01(\t\x12\x35\n\x0c\x63har_pos_map\x18\x02 \x03(\x0b\x32\x1f.CharPosSummary.CharPosMapEntry\x1a\x41\n\x0f\x43harPosMapEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1d\n\x05value\x18\x02 \x01(\x0b\x32\x0e.NumberSummary:\x02\x38\x01\"\xd7\x01\n\x0eStringsSummary\x12)\n\x0cunique_count\x18\x01 \x01(\x0b\x32\x13.UniqueCountSummary\x12)\n\x08\x66requent\x18\x02 \x01(\x0b\x32\x17.FrequentStringsSummary\x12\x1e\n\x06length\x18\x03 \x01(\x0b\x32\x0e.NumberSummary\x12$\n\x0ctoken_length\x18\x04 \x01(\x0b\x32\x0e.NumberSummary\x12)\n\x10\x63har_pos_tracker\x18\x05 \x01(\x0b\x32\x0f.CharPosSummary\"\x9d\x01\n\rSchemaSummary\x12$\n\rinferred_type\x18\x01 \x01(\x0b\x32\r.InferredType\x12\x33\n\x0btype_counts\x18\x02 \x03(\x0b\x32\x1e.SchemaSummary.TypeCountsEntry\x1a\x31\n\x0fTypeCountsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x03:\x02\x38\x01\"\x80\x01\n\x10HistogramSummary\x12\r\n\x05start\x18\x01 \x01(\x01\x12\x0b\n\x03\x65nd\x18\x02 \x01(\x01\x12\r\n\x05width\x18\x03 \x01(\x01\x12\x0e\n\x06\x63ounts\x18\x04 \x03(\x03\x12\x0b\n\x03max\x18\x05 \x01(\x01\x12\x0b\n\x03min\x18\x06 \x01(\x01\x12\x0c\n\x04\x62ins\x18\x07 \x03(\x01\x12\t\n\x01n\x18\x08 \x01(\x03\"=\n\x0fQuantileSummary\x12\x11\n\tquantiles\x18\x01 \x03(\x01\x12\x17\n\x0fquantile_values\x18\x02 \x03(\x01\"\x94\x02\n\rNumberSummary\x12\r\n\x05\x63ount\x18\x01 \x01(\x04\x12\x0b\n\x03min\x18\x02 \x01(\x01\x12\x0b\n\x03max\x18\x03 \x01(\x01\x12\x0c\n\x04mean\x18\x04 \x01(\x01\x12\x0e\n\x06stddev\x18\x05 \x01(\x01\x12$\n\thistogram\x18\x06 \x01(\x0b\x32\x11.HistogramSummary\x12)\n\x0cunique_count\x18\x07 \x01(\x0b\x32\x13.UniqueCountSummary\x12#\n\tquantiles\x18\x08 \x01(\x0b\x32\x10.QuantileSummary\x12\x31\n\x10\x66requent_numbers\x18\t \x01(\x0b\x32\x17.FrequentNumbersSummary\x12\x13\n\x0bis_discrete\x18\n \x01(\x08\"\xf9\x01\n\rColumnSummary\x12\x1d\n\x08\x63ounters\x18\x01 \x01(\x0b\x32\x0b.CountersV0\x12\x1e\n\x06schema\x18\x02 \x01(\x0b\x32\x0e.SchemaSummary\x12&\n\x0enumber_summary\x18\x03 \x01(\x0b\x32\x0e.NumberSummary\x12\'\n\x0estring_summary\x18\x04 \x01(\x0b\x32\x0f.StringsSummary\x12-\n\x0e\x66requent_items\x18\x05 \x01(\x0b\x32\x15.FrequentItemsSummary\x12)\n\x0cunique_count\x18\x06 \x01(\x0b\x32\x13.UniqueCountSummary\"\xc7\x01\n\x0e\x44\x61tasetSummary\x12(\n\nproperties\x18\x01 \x01(\x0b\x32\x14.DatasetPropertiesV0\x12-\n\x07\x63olumns\x18\x02 \x03(\x0b\x32\x1c.DatasetSummary.ColumnsEntry\x12\x1c\n\x05model\x18\x03 \x01(\x0b\x32\r.ModelSummary\x1a>\n\x0c\x43olumnsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1d\n\x05value\x18\x02 \x01(\x0b\x32\x0e.ColumnSummary:\x02\x38\x01\"0\n\x0cModelSummary\x12 \n\x07metrics\x18\x01 \x01(\x0b\x32\x0f.MetricsSummary\"\x9f\x01\n\x0eMetricsSummary\x12\x1e\n\nmodel_type\x18\x01 \x01(\x0e\x32\n.ModelType\x12\x1e\n\x0broc_fpr_tpr\x18\x02 \x01(\x0b\x32\t.ROCCurve\x12!\n\x0brecall_prec\x18\x03 \x01(\x0b\x32\x0c.RecallCurve\x12*\n\x10\x63onfusion_matrix\x18\x04 \x01(\x0b\x32\x10.ConfusionMatrix\"\x93\x01\n\x0f\x43onfusionMatrix\x12\x0e\n\x06labels\x18\x01 \x03(\t\x12\x14\n\x0ctarget_field\x18\x02 \x01(\t\x12\x19\n\x11predictions_field\x18\x03 \x01(\t\x12\x13\n\x0bscore_field\x18\x04 \x01(\t\x12*\n\x06\x63ounts\x18\x05 \x03(\x0b\x32\x1a.google.protobuf.ListValue\"6\n\x08ROCCurve\x12*\n\x06values\x18\x01 \x03(\x0b\x32\x1a.google.protobuf.ListValue\"9\n\x0bRecallCurve\x12*\n\x06values\x18\x01 \x03(\x0b\x32\x1a.google.protobuf.ListValue\"\x87\x01\n\x10\x44\x61tasetSummaries\x12\x31\n\x08profiles\x18\x01 \x03(\x0b\x32\x1f.DatasetSummaries.ProfilesEntry\x1a@\n\rProfilesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\x1e\n\x05value\x18\x02 \x01(\x0b\x32\x0f.DatasetSummary:\x02\x38\x01\x42*\n\x1b\x63om.whylogs.v0.core.messageB\tSummariesP\x01\x62\x06proto3')



_UNIQUECOUNTSUMMARY = DESCRIPTOR.message_types_by_name['UniqueCountSummary']
_FREQUENTSTRINGSSUMMARY = DESCRIPTOR.message_types_by_name['FrequentStringsSummary']
_FREQUENTSTRINGSSUMMARY_FREQUENTITEM = _FREQUENTSTRINGSSUMMARY.nested_types_by_name['FrequentItem']
_FREQUENTNUMBERSSUMMARY = DESCRIPTOR.message_types_by_name['FrequentNumbersSummary']
_FREQUENTNUMBERSSUMMARY_FREQUENTDOUBLEITEM = _FREQUENTNUMBERSSUMMARY.nested_types_by_name['FrequentDoubleItem']
_FREQUENTNUMBERSSUMMARY_FREQUENTLONGITEM = _FREQUENTNUMBERSSUMMARY.nested_types_by_name['FrequentLongItem']
_FREQUENTITEMSSUMMARY = DESCRIPTOR.message_types_by_name['FrequentItemsSummary']
_FREQUENTITEMSSUMMARY_FREQUENTITEM = _FREQUENTITEMSSUMMARY.nested_types_by_name['FrequentItem']
_CHARPOSSUMMARY = DESCRIPTOR.message_types_by_name['CharPosSummary']
_CHARPOSSUMMARY_CHARPOSMAPENTRY = _CHARPOSSUMMARY.nested_types_by_name['CharPosMapEntry']
_STRINGSSUMMARY = DESCRIPTOR.message_types_by_name['StringsSummary']
_SCHEMASUMMARY = DESCRIPTOR.message_types_by_name['SchemaSummary']
_SCHEMASUMMARY_TYPECOUNTSENTRY = _SCHEMASUMMARY.nested_types_by_name['TypeCountsEntry']
_HISTOGRAMSUMMARY = DESCRIPTOR.message_types_by_name['HistogramSummary']
_QUANTILESUMMARY = DESCRIPTOR.message_types_by_name['QuantileSummary']
_NUMBERSUMMARY = DESCRIPTOR.message_types_by_name['NumberSummary']
_COLUMNSUMMARY = DESCRIPTOR.message_types_by_name['ColumnSummary']
_DATASETSUMMARY = DESCRIPTOR.message_types_by_name['DatasetSummary']
_DATASETSUMMARY_COLUMNSENTRY = _DATASETSUMMARY.nested_types_by_name['ColumnsEntry']
_MODELSUMMARY = DESCRIPTOR.message_types_by_name['ModelSummary']
_METRICSSUMMARY = DESCRIPTOR.message_types_by_name['MetricsSummary']
_CONFUSIONMATRIX = DESCRIPTOR.message_types_by_name['ConfusionMatrix']
_ROCCURVE = DESCRIPTOR.message_types_by_name['ROCCurve']
_RECALLCURVE = DESCRIPTOR.message_types_by_name['RecallCurve']
_DATASETSUMMARIES = DESCRIPTOR.message_types_by_name['DatasetSummaries']
_DATASETSUMMARIES_PROFILESENTRY = _DATASETSUMMARIES.nested_types_by_name['ProfilesEntry']
UniqueCountSummary = _reflection.GeneratedProtocolMessageType('UniqueCountSummary', (_message.Message,), {
  'DESCRIPTOR' : _UNIQUECOUNTSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:UniqueCountSummary)
  })
_sym_db.RegisterMessage(UniqueCountSummary)

FrequentStringsSummary = _reflection.GeneratedProtocolMessageType('FrequentStringsSummary', (_message.Message,), {

  'FrequentItem' : _reflection.GeneratedProtocolMessageType('FrequentItem', (_message.Message,), {
    'DESCRIPTOR' : _FREQUENTSTRINGSSUMMARY_FREQUENTITEM,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:FrequentStringsSummary.FrequentItem)
    })
  ,
  'DESCRIPTOR' : _FREQUENTSTRINGSSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:FrequentStringsSummary)
  })
_sym_db.RegisterMessage(FrequentStringsSummary)
_sym_db.RegisterMessage(FrequentStringsSummary.FrequentItem)

FrequentNumbersSummary = _reflection.GeneratedProtocolMessageType('FrequentNumbersSummary', (_message.Message,), {

  'FrequentDoubleItem' : _reflection.GeneratedProtocolMessageType('FrequentDoubleItem', (_message.Message,), {
    'DESCRIPTOR' : _FREQUENTNUMBERSSUMMARY_FREQUENTDOUBLEITEM,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:FrequentNumbersSummary.FrequentDoubleItem)
    })
  ,

  'FrequentLongItem' : _reflection.GeneratedProtocolMessageType('FrequentLongItem', (_message.Message,), {
    'DESCRIPTOR' : _FREQUENTNUMBERSSUMMARY_FREQUENTLONGITEM,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:FrequentNumbersSummary.FrequentLongItem)
    })
  ,
  'DESCRIPTOR' : _FREQUENTNUMBERSSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:FrequentNumbersSummary)
  })
_sym_db.RegisterMessage(FrequentNumbersSummary)
_sym_db.RegisterMessage(FrequentNumbersSummary.FrequentDoubleItem)
_sym_db.RegisterMessage(FrequentNumbersSummary.FrequentLongItem)

FrequentItemsSummary = _reflection.GeneratedProtocolMessageType('FrequentItemsSummary', (_message.Message,), {

  'FrequentItem' : _reflection.GeneratedProtocolMessageType('FrequentItem', (_message.Message,), {
    'DESCRIPTOR' : _FREQUENTITEMSSUMMARY_FREQUENTITEM,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:FrequentItemsSummary.FrequentItem)
    })
  ,
  'DESCRIPTOR' : _FREQUENTITEMSSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:FrequentItemsSummary)
  })
_sym_db.RegisterMessage(FrequentItemsSummary)
_sym_db.RegisterMessage(FrequentItemsSummary.FrequentItem)

CharPosSummary = _reflection.GeneratedProtocolMessageType('CharPosSummary', (_message.Message,), {

  'CharPosMapEntry' : _reflection.GeneratedProtocolMessageType('CharPosMapEntry', (_message.Message,), {
    'DESCRIPTOR' : _CHARPOSSUMMARY_CHARPOSMAPENTRY,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:CharPosSummary.CharPosMapEntry)
    })
  ,
  'DESCRIPTOR' : _CHARPOSSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:CharPosSummary)
  })
_sym_db.RegisterMessage(CharPosSummary)
_sym_db.RegisterMessage(CharPosSummary.CharPosMapEntry)

StringsSummary = _reflection.GeneratedProtocolMessageType('StringsSummary', (_message.Message,), {
  'DESCRIPTOR' : _STRINGSSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:StringsSummary)
  })
_sym_db.RegisterMessage(StringsSummary)

SchemaSummary = _reflection.GeneratedProtocolMessageType('SchemaSummary', (_message.Message,), {

  'TypeCountsEntry' : _reflection.GeneratedProtocolMessageType('TypeCountsEntry', (_message.Message,), {
    'DESCRIPTOR' : _SCHEMASUMMARY_TYPECOUNTSENTRY,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:SchemaSummary.TypeCountsEntry)
    })
  ,
  'DESCRIPTOR' : _SCHEMASUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:SchemaSummary)
  })
_sym_db.RegisterMessage(SchemaSummary)
_sym_db.RegisterMessage(SchemaSummary.TypeCountsEntry)

HistogramSummary = _reflection.GeneratedProtocolMessageType('HistogramSummary', (_message.Message,), {
  'DESCRIPTOR' : _HISTOGRAMSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:HistogramSummary)
  })
_sym_db.RegisterMessage(HistogramSummary)

QuantileSummary = _reflection.GeneratedProtocolMessageType('QuantileSummary', (_message.Message,), {
  'DESCRIPTOR' : _QUANTILESUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:QuantileSummary)
  })
_sym_db.RegisterMessage(QuantileSummary)

NumberSummary = _reflection.GeneratedProtocolMessageType('NumberSummary', (_message.Message,), {
  'DESCRIPTOR' : _NUMBERSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:NumberSummary)
  })
_sym_db.RegisterMessage(NumberSummary)

ColumnSummary = _reflection.GeneratedProtocolMessageType('ColumnSummary', (_message.Message,), {
  'DESCRIPTOR' : _COLUMNSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:ColumnSummary)
  })
_sym_db.RegisterMessage(ColumnSummary)

DatasetSummary = _reflection.GeneratedProtocolMessageType('DatasetSummary', (_message.Message,), {

  'ColumnsEntry' : _reflection.GeneratedProtocolMessageType('ColumnsEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETSUMMARY_COLUMNSENTRY,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:DatasetSummary.ColumnsEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:DatasetSummary)
  })
_sym_db.RegisterMessage(DatasetSummary)
_sym_db.RegisterMessage(DatasetSummary.ColumnsEntry)

ModelSummary = _reflection.GeneratedProtocolMessageType('ModelSummary', (_message.Message,), {
  'DESCRIPTOR' : _MODELSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:ModelSummary)
  })
_sym_db.RegisterMessage(ModelSummary)

MetricsSummary = _reflection.GeneratedProtocolMessageType('MetricsSummary', (_message.Message,), {
  'DESCRIPTOR' : _METRICSSUMMARY,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:MetricsSummary)
  })
_sym_db.RegisterMessage(MetricsSummary)

ConfusionMatrix = _reflection.GeneratedProtocolMessageType('ConfusionMatrix', (_message.Message,), {
  'DESCRIPTOR' : _CONFUSIONMATRIX,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:ConfusionMatrix)
  })
_sym_db.RegisterMessage(ConfusionMatrix)

ROCCurve = _reflection.GeneratedProtocolMessageType('ROCCurve', (_message.Message,), {
  'DESCRIPTOR' : _ROCCURVE,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:ROCCurve)
  })
_sym_db.RegisterMessage(ROCCurve)

RecallCurve = _reflection.GeneratedProtocolMessageType('RecallCurve', (_message.Message,), {
  'DESCRIPTOR' : _RECALLCURVE,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:RecallCurve)
  })
_sym_db.RegisterMessage(RecallCurve)

DatasetSummaries = _reflection.GeneratedProtocolMessageType('DatasetSummaries', (_message.Message,), {

  'ProfilesEntry' : _reflection.GeneratedProtocolMessageType('ProfilesEntry', (_message.Message,), {
    'DESCRIPTOR' : _DATASETSUMMARIES_PROFILESENTRY,
    '__module__' : 'v0_summaries_pb2'
    # @@protoc_insertion_point(class_scope:DatasetSummaries.ProfilesEntry)
    })
  ,
  'DESCRIPTOR' : _DATASETSUMMARIES,
  '__module__' : 'v0_summaries_pb2'
  # @@protoc_insertion_point(class_scope:DatasetSummaries)
  })
_sym_db.RegisterMessage(DatasetSummaries)
_sym_db.RegisterMessage(DatasetSummaries.ProfilesEntry)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033com.whylogs.v0.core.messageB\tSummariesP\001'
  _CHARPOSSUMMARY_CHARPOSMAPENTRY._options = None
  _CHARPOSSUMMARY_CHARPOSMAPENTRY._serialized_options = b'8\001'
  _SCHEMASUMMARY_TYPECOUNTSENTRY._options = None
  _SCHEMASUMMARY_TYPECOUNTSENTRY._serialized_options = b'8\001'
  _DATASETSUMMARY_COLUMNSENTRY._options = None
  _DATASETSUMMARY_COLUMNSENTRY._serialized_options = b'8\001'
  _DATASETSUMMARIES_PROFILESENTRY._options = None
  _DATASETSUMMARIES_PROFILESENTRY._serialized_options = b'8\001'
  _UNIQUECOUNTSUMMARY._serialized_start=71
  _UNIQUECOUNTSUMMARY._serialized_end=139
  _FREQUENTSTRINGSSUMMARY._serialized_start=141
  _FREQUENTSTRINGSSUMMARY._serialized_end=267
  _FREQUENTSTRINGSSUMMARY_FREQUENTITEM._serialized_start=220
  _FREQUENTSTRINGSSUMMARY_FREQUENTITEM._serialized_end=267
  _FREQUENTNUMBERSSUMMARY._serialized_start=270
  _FREQUENTNUMBERSSUMMARY._serialized_end=548
  _FREQUENTNUMBERSSUMMARY_FREQUENTDOUBLEITEM._serialized_start=414
  _FREQUENTNUMBERSSUMMARY_FREQUENTDOUBLEITEM._serialized_end=481
  _FREQUENTNUMBERSSUMMARY_FREQUENTLONGITEM._serialized_start=483
  _FREQUENTNUMBERSSUMMARY_FREQUENTLONGITEM._serialized_end=548
  _FREQUENTITEMSSUMMARY._serialized_start=550
  _FREQUENTITEMSSUMMARY._serialized_end=677
  _FREQUENTITEMSSUMMARY_FREQUENTITEM._serialized_start=625
  _FREQUENTITEMSSUMMARY_FREQUENTITEM._serialized_end=677
  _CHARPOSSUMMARY._serialized_start=680
  _CHARPOSSUMMARY._serialized_end=842
  _CHARPOSSUMMARY_CHARPOSMAPENTRY._serialized_start=777
  _CHARPOSSUMMARY_CHARPOSMAPENTRY._serialized_end=842
  _STRINGSSUMMARY._serialized_start=845
  _STRINGSSUMMARY._serialized_end=1060
  _SCHEMASUMMARY._serialized_start=1063
  _SCHEMASUMMARY._serialized_end=1220
  _SCHEMASUMMARY_TYPECOUNTSENTRY._serialized_start=1171
  _SCHEMASUMMARY_TYPECOUNTSENTRY._serialized_end=1220
  _HISTOGRAMSUMMARY._serialized_start=1223
  _HISTOGRAMSUMMARY._serialized_end=1351
  _QUANTILESUMMARY._serialized_start=1353
  _QUANTILESUMMARY._serialized_end=1414
  _NUMBERSUMMARY._serialized_start=1417
  _NUMBERSUMMARY._serialized_end=1693
  _COLUMNSUMMARY._serialized_start=1696
  _COLUMNSUMMARY._serialized_end=1945
  _DATASETSUMMARY._serialized_start=1948
  _DATASETSUMMARY._serialized_end=2147
  _DATASETSUMMARY_COLUMNSENTRY._serialized_start=2085
  _DATASETSUMMARY_COLUMNSENTRY._serialized_end=2147
  _MODELSUMMARY._serialized_start=2149
  _MODELSUMMARY._serialized_end=2197
  _METRICSSUMMARY._serialized_start=2200
  _METRICSSUMMARY._serialized_end=2359
  _CONFUSIONMATRIX._serialized_start=2362
  _CONFUSIONMATRIX._serialized_end=2509
  _ROCCURVE._serialized_start=2511
  _ROCCURVE._serialized_end=2565
  _RECALLCURVE._serialized_start=2567
  _RECALLCURVE._serialized_end=2624
  _DATASETSUMMARIES._serialized_start=2627
  _DATASETSUMMARIES._serialized_end=2762
  _DATASETSUMMARIES_PROFILESENTRY._serialized_start=2698
  _DATASETSUMMARIES_PROFILESENTRY._serialized_end=2762
# @@protoc_insertion_point(module_scope)



from abc import ABC, abstractmethod
from logging import getLogger
from typing import Any, Dict, List, Optional

from whylogs.api.reader import Reader, Readers
from whylogs.api.writer import Writer, Writers
from whylogs.api.writer.writer import Writable
from whylogs.core import DatasetProfile, DatasetProfileView, Segment
from whylogs.core.metrics.metrics import Metric
from whylogs.core.model_performance_metrics import ModelPerformanceMetrics
from whylogs.core.segmentation_partition import SegmentationPartition
from whylogs.core.view.segmented_dataset_profile_view import SegmentedDatasetProfileView

logger = getLogger(__name__)


class ResultSetWriter:
    """
    Result of a logging call.
    A result set might contain one or multiple profiles or profile views.
    """

    def __init__(self, results: "ResultSet", writer: Writer):
        self._result_set = results
        self._writer = writer

    def option(self, **kwargs: Any) -> "ResultSetWriter":
        self._writer.option(**kwargs)
        return self

    def write(self, **kwargs: Any) -> None:
        # multi-profile writer
        files = self._result_set.get_writables()
        if not files:
            logger.warning("Attempt to write a result set with no writables returned, nothing written!")
        else:
            logger.debug(f"About to write {len(files)} files:")
            # TODO: special handling of large number of files, handle throttling
            for view in files:
                self._writer.write(profile=view, **kwargs)
            logger.debug(f"Completed writing {len(files)} files!")


class ResultSetReader:
    def __init__(self, reader: Reader) -> None:
        self._reader = reader

    def option(self, **kwargs: Any) -> "ResultSetReader":
        self._reader.option(**kwargs)
        return self

    def read(self, **kwargs: Any) -> "ResultSet":
        return self._reader.read(**kwargs)


class ResultSet(ABC):
    """
    A holder object for profiling results.

    A whylogs.log call can result in more than one profile. This wrapper class
    simplifies the navigation among these profiles.

    Note that currently we only hold one profile but we're planning to add other
    kinds of profiles such as segmented profiles here.
    """

    @staticmethod
    def read(multi_profile_file: str) -> "ResultSet":
        # TODO: parse multiple profile
        view = DatasetProfileView.read(multi_profile_file)
        return ViewResultSet(view=view)

    @staticmethod
    def reader(name: str = "local") -> "ResultSetReader":
        reader = Readers.get(name)
        return ResultSetReader(reader=reader)

    def writer(self, name: str = "local") -> "ResultSetWriter":
        writer = Writers.get(name)
        return ResultSetWriter(results=self, writer=writer)

    @abstractmethod
    def view(self) -> Optional[DatasetProfileView]:
        pass

    @abstractmethod
    def profile(self) -> Optional[DatasetProfile]:
        pass

    def get_writables(self) -> Optional[List[Writable]]:
        return [self.view()]

    @property
    def performance_metrics(self) -> Optional[ModelPerformanceMetrics]:
        profile = self.profile()
        if profile:
            return profile.model_performance_metrics
        return None

    def add_model_performance_metrics(self, metrics: ModelPerformanceMetrics) -> None:
        profile = self.profile()
        if profile:
            profile.add_model_performance_metrics(metrics)
        else:
            raise ValueError("Cannot add performance metrics to a result set with no profile!")

    def add_metrics(self, name: str, metric: Metric) -> None:
        profile = self.profile()
        if profile:
            profile.add_dataset_metric(name, metric)
        else:
            raise ValueError(f"Cannot add {name} metric {metric} to a result set with no profile!")


class ViewResultSet(ResultSet):
    def __init__(self, view: DatasetProfileView) -> None:
        self._view = view

    def profile(self) -> Optional[DatasetProfile]:
        raise ValueError("No profile available. Can only view")

    def view(self) -> Optional[DatasetProfileView]:
        return self._view


class ProfileResultSet(ResultSet):
    def __init__(self, profile: DatasetProfile) -> None:
        self._profile = profile

    def profile(self) -> Optional[DatasetProfile]:
        return self._profile

    def view(self) -> Optional[DatasetProfileView]:
        return self._profile.view()


class SegmentedResultSet(ResultSet):
    def __init__(
        self,
        segments: Dict[str, Dict[Segment, DatasetProfile]],
        partitions: Optional[List[SegmentationPartition]] = None,
    ) -> None:
        self._segments = segments
        self._partitions = partitions
        self.model_performance_metric: Optional[ModelPerformanceMetrics] = None

    def profile(self, segment: Optional[Segment] = None) -> Optional[DatasetProfile]:
        if not self._segments:
            return None
        elif segment:
            paritition_segments = self._segments.get(segment.parent_id)
            return paritition_segments.get(segment) if paritition_segments else None
        # special case return a single segment if there is only one, even if not specified
        elif len(self._segments) == 1:
            for partition_id in self._segments:
                segments = self._segments.get(partition_id)
                number_of_segments = len(segments) if segments else 0
                if number_of_segments == 1:
                    single_dictionary: Dict[Segment, DatasetProfile] = segments if segments else dict()
                    for key in single_dictionary:
                        return single_dictionary[key]

        raise ValueError(
            f"A profile was requested from a segmented result set without specifying which segment to return: {self._segments}"
        )

    @property
    def partitions(self) -> Optional[List[SegmentationPartition]]:
        return self._partitions

    def segments(self, restrict_to_parition_id: Optional[str] = None) -> Optional[List[Segment]]:
        result: Optional[List[Segment]] = None
        if not self._segments:
            return result
        result = list()
        if restrict_to_parition_id:
            segments = self._segments.get(restrict_to_parition_id)
            if segments:
                for segment in segments:
                    result.append(segment)
        else:
            for partition_id in self._segments:
                for segment in self._segments[partition_id]:
                    result.append(segment)
        return result

    @property
    def count(self) -> int:
        result = 0
        if self._segments:
            for segment_key in self._segments:
                profiles = self._segments[segment_key]
                result += len(profiles)
        return result

    def segments_in_partition(self, partition: SegmentationPartition) -> Optional[Dict[Segment, DatasetProfile]]:
        return self._segments.get(partition.id)

    def view(self, segment: Optional[Segment] = None) -> Optional[DatasetProfileView]:
        result = self.profile(segment)
        return result.view() if result else None

    def get_model_performance_metrics_for_segment(self, segment: Segment) -> Optional[ModelPerformanceMetrics]:
        if segment.parent_id in self._segments:
            profile = self._segments[segment.parent_id][segment]
            if not profile:
                logger.warning(
                    f"No profile found for segment {segment} when requesting model performance metrics, returning None!"
                )
                return None
            view = profile.view()
            return view.model_performance_metrics
        return None

    def get_writables(self) -> Optional[List[Writable]]:
        results: Optional[List[Writable]] = None
        if self._segments:
            results = []
            logger.info(f"Building list of: {self.count} SegmentedDatasetProfileViews in SegmentedResultSet.")
            # TODO: handle more than one partition
            if not self.partitions:
                raise ValueError(
                    f"Building list of: {self.count} SegmentedDatasetProfileViews in SegmentedResultSet but no partitions found: {self.partitions}."
                )
            if len(self.partitions) > 1:
                logger.error(
                    f"Building list of: {self.count} SegmentedDatasetProfileViews in SegmentedResultSet but found more than one partition: "
                    f"{self.partitions}. Using first partition only!!"
                )
            first_partition = self.partitions[0]
            segments = self.segments_in_partition(first_partition)
            if segments:
                for segment_key in segments:
                    profile = segments[segment_key]
                    metric = self.get_model_performance_metrics_for_segment(segment_key)
                    if metric:
                        profile.add_model_performance_metrics(metric)
                        logger.debug(
                            f"Found model performance metrics: {metric}, adding to segmented profile: {segment_key}."
                        )

                    segmented_profile = SegmentedDatasetProfileView(
                        profile_view=profile.view(), segment=segment_key, partition=first_partition
                    )
                    results.append(segmented_profile)
            else:
                logger.warning(
                    f"Found no segments in partition: {first_partition} even though we have: {self.count} segments overall"
                )
            logger.info(f"From list of: {self.count} SegmentedDatasetProfileViews using {len(results)}")
        else:
            logger.warning(
                f"Attempt to build segmented results for writing but there are no segments in this result set: {self._segments}. returning None."
            )
        return results

    def add_metrics_for_segment(self, metrics: ModelPerformanceMetrics, segment: Segment) -> None:
        if segment.parent_id in self._segments:
            profile = self._segments[segment.parent_id][segment]
            profile.add_model_performance_metrics(metrics)



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, TypeVar

if TYPE_CHECKING:
    from whylogs import ResultSet

import whylogs as why

T = TypeVar("T", bound="Reader")


class Reader(ABC):
    @abstractmethod
    def read(self, **kwargs: Any) -> "ResultSet":
        """
        Must be implemented by all inherited Readers, declaring
        how to fetch files from their locations to a local temp dir
        """

    @abstractmethod
    def option(self: T, **kwargs: Any) -> T:
        pass

    @staticmethod
    def get_file_from_path(path) -> "ResultSet":
        return why.read(path)



import logging
from typing import Optional

from whylogs import DatasetProfileView
from whylogs.viz.enums.enums import PageSpecEnum
from whylogs.viz.extensions.reports.html_report import HTMLReport
from whylogs.viz.utils.html_template_utils import get_compiled_template
from whylogs.viz.utils.profile_viz_calculations import generate_profile_summary

logger = logging.getLogger(__name__)


class ProfileSummaryReport(HTMLReport):
    def __init__(self, target_view: DatasetProfileView, height: Optional[str] = None):
        super().__init__(target_view=target_view, height=height)

    def report(self) -> str:
        page_spec = PageSpecEnum.PROFILE_SUMMARY.value
        template = get_compiled_template(page_spec.html)

        profile_summary = generate_profile_summary(self.target_view, config=None)
        rendered_template = template(profile_summary)
        profile_summary_report = self.display(rendered_template, page_spec)
        return profile_summary_report



from typing import Union

from whylogs.core.metrics.metrics import CardinalityMetric

from ..metric_constraints import MetricConstraint, MetricsSelector


def distinct_number_in_range(column_name: str, lower: Union[int, float], upper: Union[int, float]) -> MetricConstraint:
    """Number of distinct categories must be between lower and upper values (inclusive).

    Parameters
    ----------
    column_name : str
        Column the constraint is applied to
    lower : int
        Lower bound of defined range
    upper : int
        Upper bound of the value range
    """

    def distinct_in_range(metric: CardinalityMetric):
        distinct_estimate = metric.hll.value.get_estimate()
        if not lower > distinct_estimate and not distinct_estimate > upper:
            return True
        return False

    constraint = MetricConstraint(
        name=f"{column_name} distinct values estimate between {lower} and {upper} (inclusive)",
        condition=distinct_in_range,
        metric_selector=MetricsSelector(column_name=column_name, metric_name="cardinality"),
    )
    return constraint



from typing import List, Union

from whylogs.core.stubs import np as np
from whylogs.core.stubs import pd as pd

COMMON_COLUMNAR_TYPES = Union[pd.Series, np.ndarray, List]
LARGE_CACHE_SIZE_LIMIT = 1024 * 100



import logging
from typing import Any, Dict, List

from whylogs.core.metrics import Metric
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.projectors import FieldProjector
from whylogs.core.proto import ColumnMessage
from whylogs.core.schema import ColumnSchema
from whylogs.core.view import ColumnProfileView

logger = logging.getLogger(__name__)


class ColumnProfile(object):
    def __init__(self, name: str, schema: ColumnSchema, cache_size: int):
        self._name = name
        self._schema = schema

        self._metrics = schema.get_metrics(name)
        self._column_validators = schema.get_validators(name)
        self._projector = FieldProjector(col_names=[name])
        self._success_count = 0
        self._failure_count = 0
        logger.debug("Setting cache size for column: %s with size: %s", self._name, cache_size)
        self._cache_size = cache_size
        self._cache: List[Any] = []

    def add_metric(self, metric: Metric) -> None:
        if metric.namespace in self._metrics:
            logger.warning(f"Replacing metric {metric.namespace} in column {self._name}")

        self._metrics[metric.namespace] = metric

    def track(self, row: Dict[str, Any]) -> None:
        value = self._projector.apply(row)
        if len(self._cache) < self._cache_size - 1:
            self._cache.append(value)
        else:
            self._cache.append(value)
            cache_list = self._cache
            self._cache = []
            self.track_column(cache_list)

    def flush(self) -> None:
        """Force emptying the cache and update the internal metrics."""

        logger.debug("Flushing out the cache for col: %s. Cache size: %s", self._name, self._cache_size)
        old_cache = self._cache
        self._cache = []
        self.track_column(old_cache)

    def track_column(self, series: Any) -> None:
        ex_col = PreprocessedColumn.apply(series)
        for metric in self._metrics.values():
            res = metric.columnar_update(ex_col)
            self._failure_count += res.failures
            self._success_count += res.successes
        for validator in self._column_validators:
            validator.columnar_validate(ex_col)

    def to_protobuf(self) -> ColumnMessage:
        self.flush()

        return self.view().to_protobuf()

    def view(self) -> ColumnProfileView:
        self.flush()

        return ColumnProfileView(
            metrics=self._metrics.copy(),
            success_count=self._success_count,
            failure_count=self._failure_count,
        )



import logging
from typing import Any, Dict, Iterable, Iterator, List, Mapping, Optional, Tuple

from whylogs.api.logger.result_set import SegmentedResultSet
from whylogs.api.store.profile_store import ProfileStore
from whylogs.core import DatasetSchema
from whylogs.core.dataset_profile import DatasetProfile
from whylogs.core.input_resolver import _pandas_or_dict
from whylogs.core.segment import Segment
from whylogs.core.segmentation_partition import SegmentationPartition, SegmentFilter
from whylogs.core.stubs import pd

logger = logging.getLogger(__name__)

_MAX_SEGMENT_PARTITIONS = 10


def _process_segment(
    segmented_data: Any,
    segment_key: Segment,
    segments: Dict[Segment, Any],
    schema: DatasetSchema,
    store: Optional[ProfileStore] = None,
):
    profile = None
    if store:
        profile = store.get_matching_profiles(segmented_data, segment=segment_key)

    if profile is None:
        profile = DatasetProfile(schema)

    profile.track(segmented_data)
    segments[segment_key] = profile


def _process_simple_partition(
    partition_id: str,
    schema: DatasetSchema,
    segments: Dict[Segment, Any],
    columns: List[str],
    pandas: Optional[pd.DataFrame] = None,
    row: Optional[Mapping[str, Any]] = None,
    profile_cache: Optional[ProfileStore] = None,
):
    if pandas is not None:
        # simple means we can segment on column values
        grouped_data = pandas.groupby(columns)
        for group in grouped_data.groups.keys():
            pandas_segment = grouped_data.get_group(group)
            if isinstance(group, str):
                segment_tuple_key: Tuple[str, ...] = (group,)
            elif isinstance(group, (List, Iterable, Iterator)):
                segment_tuple_key = tuple(str(k) for k in group)
            else:
                segment_tuple_key = (str(group),)
            segment_key = Segment(segment_tuple_key, partition_id)
            _process_segment(pandas_segment, segment_key, segments, schema, profile_cache)
    elif row:
        # TODO: consider if we need to combine with the column names
        segment_key = Segment(tuple(str(row[element]) for element in columns), partition_id)
        _process_segment(row, segment_key, segments, schema, profile_cache)


def _filter_inputs(
    filter: SegmentFilter, pandas: Optional[pd.DataFrame] = None, row: Optional[Mapping[str, Any]] = None
) -> Tuple[Optional[pd.DataFrame], Optional[Dict[str, Any]]]:
    assert (
        filter.filter_function or filter.query_string
    ), f"must define at least a filter function or query string when specifying a segment filter: {filter}"
    filtered_pandas = None
    filtered_row = None
    if pandas is not None:
        if filter.filter_function:
            filtered_pandas = pandas[filter.filter_function]
        elif filter.query_string:
            filtered_pandas = pandas.query(filter.query_string)
    elif row is not None:
        if filter.filter_function:
            filtered_row = filter.filter_function(row)
        elif filter.query_string:
            raise ValueError(
                "SegmentFilter query string not supported when logging rows, either don't specify a filter or implement the filter.filter_function"
            )
    return (filtered_pandas, filtered_row)


def _log_segment(
    partition: SegmentationPartition,
    schema: DatasetSchema,
    obj: Any = None,
    pandas: Optional[pd.DataFrame] = None,
    row: Optional[Mapping[str, Any]] = None,
    store: Optional[ProfileStore] = None,
) -> Dict[Segment, Any]:
    segments: Dict[Segment, Any] = {}
    pandas, row = _pandas_or_dict(obj, pandas, row)
    if partition.filter:
        pandas, row = _filter_inputs(partition.filter, pandas, row)
    if partition.simple:
        columns = partition.mapper.col_names if partition.mapper else None
        if columns:
            _process_simple_partition(partition.id, schema, segments, columns, pandas, row, store)
    else:
        raise NotImplementedError("custom mapped segments not yet implemented")
    return segments


def segment_processing(
    schema: DatasetSchema,
    obj: Any = None,
    pandas: Optional[pd.DataFrame] = None,
    row: Optional[Dict[str, Any]] = None,
) -> SegmentedResultSet:
    number_of_partitions = len(schema.segments)
    logger.info(f"The specified schema defines segments with {number_of_partitions} partitions.")
    assert not (
        number_of_partitions > _MAX_SEGMENT_PARTITIONS
    ), f"Attempt to process {number_of_partitions} partitions is larger than the max of {_MAX_SEGMENT_PARTITIONS}, use a lower number of partitions"
    segmented_profiles = dict()
    segment_partitions = list()

    for partition_name in schema.segments:
        segment_partition = schema.segments[partition_name]
        logger.info(f"Processing partition with name({partition_name})")
        logger.debug(f"{partition_name}: is simple ({segment_partition.simple}), id ({segment_partition.id})")
        if segment_partition.filter:
            # TODO segments filter
            logger.debug(f"{partition_name}: defines filter ({segment_partition.filter})")
        if segment_partition.mapper:
            logger.debug(
                f"{partition_name}: defines mapper on colums ({segment_partition.mapper.col_names}) and id ({segment_partition.mapper.id})"
            )
        partition_segments = _log_segment(segment_partition, schema, obj, pandas, row)
        segmented_profiles[segment_partition.id] = partition_segments
        segment_partitions.append(segment_partition)
        logger.debug(f"Done profiling for partition with name({partition_name})")
    return SegmentedResultSet(segments=segmented_profiles, partitions=segment_partitions)



from typing import Any, Dict, List


class FieldProjector(object):
    def __init__(self, col_names: List[str]) -> None:
        self._col_names = col_names

    def apply(self, row: Dict[str, Any]) -> Dict[str, Any]:
        row_projection = {col_name: row.get(col_name) for col_name in self._col_names}
        return row_projection



import logging
import re
from copy import copy
from dataclasses import dataclass, field
from enum import Enum
from itertools import chain
from typing import Any, Callable, Dict, List, Set, Union

from whylogs.core.configs import SummaryConfig
from whylogs.core.metrics.metric_components import IntegralComponent, MetricComponent
from whylogs.core.metrics.metrics import Metric, MetricConfig, OperationResult
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.proto import MetricMessage

logger = logging.getLogger(__name__)


class Relation(Enum):
    match = 1
    fullmatch = 2
    equal = 3
    less = 4
    leq = 5
    greater = 6
    geq = 7
    neq = 8


def relation(op: Relation, value: Union[str, int, float]) -> Callable[[Any], bool]:
    if op == Relation.match:
        return lambda x: re.compile(value).match(x)  # type: ignore
    if op == Relation.fullmatch:
        return lambda x: re.compile(value).fullmatch(x)  # type: ignore
    if op == Relation.equal:
        return lambda x: x == value  # type: ignore
    if op == Relation.less:
        return lambda x: x < value  # type: ignore
    if op == Relation.leq:
        return lambda x: x <= value  # type: ignore
    if op == Relation.greater:
        return lambda x: x > value  # type: ignore
    if op == Relation.geq:
        return lambda x: x >= value  # type: ignore
    if op == Relation.neq:
        return lambda x: x != value  # type: ignore
    raise ValueError("Unknown ConditionCountMetric predicate")


def and_relations(left: Callable[[Any], bool], right: Callable[[Any], bool]) -> Callable[[Any], bool]:
    return lambda x: left(x) and right(x)


def or_relations(left: Callable[[Any], bool], right: Callable[[Any], bool]) -> Callable[[Any], bool]:
    return lambda x: left(x) or right(x)


def not_relation(relation: Callable[[Any], bool]) -> Callable[[Any], bool]:
    return lambda x: not relation(x)


@dataclass(frozen=True)
class Condition:
    relation: Callable[[Any], bool]
    throw_on_failure: bool = False
    log_on_failure: bool = False


@dataclass(frozen=True)
class ConditionCountConfig(MetricConfig):
    conditions: Dict[str, Condition] = field(default_factory=dict)


@dataclass(frozen=True)
class ConditionCountMetric(Metric):
    conditions: Dict[str, Condition]
    total: IntegralComponent
    matches: Dict[str, IntegralComponent] = field(default_factory=dict)

    @property
    def namespace(self) -> str:
        return "condition_count"

    def __post_init__(self) -> None:
        super(type(self), self).__post_init__()
        if "total" in self.conditions.keys():
            raise ValueError("Condition cannot be named 'total'")

        for cond_name in self.conditions.keys():
            if cond_name not in self.matches:
                self.matches[cond_name] = IntegralComponent(0)

    def merge(self, other: "ConditionCountMetric") -> "ConditionCountMetric":
        if set(self.matches.keys()) != set(other.matches.keys()):
            # log warning?
            matches = {cond_name: IntegralComponent(comp.value) for cond_name, comp in self.matches.items()}
            total = self.total.value
        else:
            matches = {
                cond_name: IntegralComponent(self.matches[cond_name].value + other.matches[cond_name].value)
                for cond_name in self.matches.keys()
            }
            total = self.total.value + other.total.value

        return ConditionCountMetric(copy(self.conditions), IntegralComponent(total), matches)

    def add_conditions(self, conditions: Dict[str, Condition]) -> None:
        if "total" in conditions.keys():
            raise ValueError("Condition cannot be named 'total'")
        for cond_name, cond in conditions.items():
            self.conditions[cond_name] = cond
            self.matches[cond_name] = IntegralComponent(0)

    def get_component_paths(self) -> List[str]:
        paths: List[str] = [
            "total",
        ] + list(self.conditions.keys())
        return paths

    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        if data.len <= 0:
            return OperationResult.ok(0)

        count = 0
        failed_conditions: Set[str] = set()
        for x in list(chain.from_iterable(data.raw_iterator())):
            count += 1
            for cond_name, condition in self.conditions.items():
                try:
                    if condition.relation(x):
                        self.matches[cond_name].set(self.matches[cond_name].value + 1)
                    else:
                        failed_conditions.add(cond_name)

                except:  # noqa
                    pass

        self.total.set(self.total.value + count)
        if condition.log_on_failure:
            logger.warning(f"Condition(s) {', '.join(failed_conditions)} failed")
        if condition.throw_on_failure:
            raise ValueError(f"Condition(s) {', '.join(failed_conditions)} failed")
        return OperationResult.ok(count)

    @classmethod
    def zero(cls, config: MetricConfig) -> "ConditionCountMetric":
        if config is None or not isinstance(config, ConditionCountConfig):
            raise ValueError("ConditionCountMetric.zero() requires ConditionCountConfig argument")

        return ConditionCountMetric(
            conditions=copy(config.conditions),
            total=IntegralComponent(0),
        )

    def to_protobuf(self) -> MetricMessage:
        msg = {"total": self.total.to_protobuf()}
        for cond_name in self.conditions.keys():
            msg[cond_name] = self.matches[cond_name].to_protobuf()

        return MetricMessage(metric_components=msg)

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        summary = {"total": self.total.value}
        for cond_name in self.matches.keys():
            summary[cond_name] = self.matches[cond_name].value

        return summary

    @classmethod
    def from_protobuf(cls, msg: MetricMessage) -> "ConditionCountMetric":
        cond_names: Set[str] = set(msg.metric_components.keys())
        cond_names.remove("total")

        conditions = {cond_name: lambda x: False for cond_name in cond_names}
        total = MetricComponent.from_protobuf(msg.metric_components["total"])
        matches = {
            cond_name: MetricComponent.from_protobuf(msg.metric_components[cond_name]) for cond_name in cond_names
        }
        return ConditionCountMetric(
            conditions,
            total,
            matches,
        )



import logging
import tempfile
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Optional

from google.protobuf.message import DecodeError

from whylogs.api.writer.writer import Writable
from whylogs.core.configs import SummaryConfig
from whylogs.core.errors import DeserializationError
from whylogs.core.proto import (
    ChunkHeader,
    ChunkMessage,
    ChunkOffsets,
    ColumnMessage,
    DatasetProfileHeader,
    DatasetProperties,
    DatasetSegmentHeader,
    MetricComponentMessage,
)
from whylogs.core.stubs import pd
from whylogs.core.utils import read_delimited_protobuf, write_delimited_protobuf
from whylogs.core.utils.timestamp_calculations import to_utc_milliseconds
from whylogs.core.view.column_profile_view import ColumnProfileView

# Magic header for whylogs using the first 4 bytes
WHYLOGS_MAGIC_HEADER = "WHY1"
WHYLOGS_MAGIC_HEADER_LEN = 4

WHYLOGS_MAGIC_HEADER_BYTES = WHYLOGS_MAGIC_HEADER.encode("utf-8")
_MODEL_PERFORMANCE = "model_performance_metrics"

logger = logging.getLogger(__name__)


class SummaryType(str, Enum):
    COLUMN = "COLUMN"
    DATASET = "DATASET"


class DatasetProfileView(Writable):
    _columns: Dict[str, ColumnProfileView]

    def __init__(
        self,
        *,
        columns: Dict[str, ColumnProfileView],
        dataset_timestamp: datetime,
        creation_timestamp: datetime,
        metrics: Optional[Dict[str, Any]] = None,
    ):
        self._columns = columns.copy()
        self._dataset_timestamp = dataset_timestamp
        self._creation_timestamp = creation_timestamp
        self._metrics = metrics

    @property
    def dataset_timestamp(self) -> datetime:
        return self._dataset_timestamp

    @property
    def creation_timestamp(self) -> datetime:
        return self._creation_timestamp

    @property
    def model_performance_metrics(self) -> Any:
        if self._metrics:
            return self._metrics.get(_MODEL_PERFORMANCE)
        return None

    def merge(self, other: "DatasetProfileView") -> "DatasetProfileView":
        all_names = set(self._columns.keys()).union(other._columns.keys())
        merged_columns: Dict[str, ColumnProfileView] = {}
        for n in all_names:
            lhs = self._columns.get(n)
            rhs = other._columns.get(n)

            res = lhs
            if lhs is None:
                res = rhs
            elif rhs is not None:
                res = lhs + rhs
            assert res is not None
            merged_columns[n] = res
        return DatasetProfileView(
            columns=merged_columns,
            dataset_timestamp=self._dataset_timestamp if self._dataset_timestamp else other.dataset_timestamp,
            creation_timestamp=self._creation_timestamp if self._creation_timestamp else other.creation_timestamp,
        )

    def get_column(self, col_name: str) -> Optional[ColumnProfileView]:
        return self._columns.get(col_name)

    def get_columns(self, col_names: Optional[List[str]] = None) -> Dict[str, ColumnProfileView]:
        if col_names:
            return {k: self._columns.get(k) for k in col_names}
        else:
            return {k: self._columns.get(k) for k in self._columns}

    def get_default_path(self) -> str:
        return f"profile_{self.creation_timestamp}.bin"

    def write(self, path: Optional[str] = None, **kwargs: Any) -> None:
        path = path or self.get_default_path()
        if self._metrics and _MODEL_PERFORMANCE in self._metrics:
            from whylogs.migration.converters import v1_to_dataset_profile_message_v0

            message_v0 = v1_to_dataset_profile_message_v0(self, None, None)
            with open(path, "w+b") as out_f:
                write_delimited_protobuf(out_f, message_v0)

            return

        all_metric_component_names = set()

        # capture the list of all metric component paths
        for col in self._columns.values():
            all_metric_component_names.update(col.get_metric_component_paths())
        metric_name_list = list(all_metric_component_names)
        metric_name_list.sort()
        metric_name_indices: Dict[str, int] = {}
        metric_index_to_name: Dict[int, str] = {}
        for i in range(0, len(metric_name_list)):
            metric_name_indices[metric_name_list[i]] = i
            metric_index_to_name[i] = metric_name_list[i]

        column_chunk_offsets: Dict[str, ChunkOffsets] = {}
        with tempfile.TemporaryFile("w+b") as f:
            for col_name in sorted(self._columns.keys()):
                column_chunk_offsets[col_name] = ChunkOffsets(offsets=[f.tell()])

                col = self._columns[col_name]

                # for a given column, turn it into a ChunkMessage.
                indexed_component_messages: Dict[int, MetricComponentMessage] = {}
                metric_components = col.to_protobuf().metric_components
                for m_name, m_component in metric_components.items():
                    index = metric_name_indices.get(m_name)
                    if index is None:
                        raise ValueError(f"Missing metric from index mapping. Metric name: {m_name}")
                    indexed_component_messages[index] = m_component

                chunk_msg = ChunkMessage(metric_components=indexed_component_messages)
                chunk_header = ChunkHeader(type=ChunkHeader.ChunkType.COLUMN, length=chunk_msg.ByteSize())
                write_delimited_protobuf(f, chunk_header)
                f.write(chunk_msg.SerializeToString())

            total_len = f.tell()
            f.flush()

            properties = DatasetProperties(
                dataset_timestamp=to_utc_milliseconds(self._dataset_timestamp),
                creation_timestamp=to_utc_milliseconds(self._creation_timestamp),
            )
            dataset_header = DatasetProfileHeader(
                column_offsets=column_chunk_offsets,
                properties=properties,
                length=total_len,
                indexed_metric_paths=metric_index_to_name,
            )

            # single file segments
            dataset_segment_header = DatasetSegmentHeader(
                has_segments=False,
            )

            with open(path, "w+b") as out_f:
                out_f.write(WHYLOGS_MAGIC_HEADER_BYTES)
                write_delimited_protobuf(out_f, dataset_segment_header)
                write_delimited_protobuf(out_f, dataset_header)

                f.seek(0)
                while f.tell() < total_len:
                    buffer = f.read(1024)
                    out_f.write(buffer)

    @classmethod
    def read(cls, path: str) -> "DatasetProfileView":
        with open(path, "r+b") as f:
            buf = f.read(WHYLOGS_MAGIC_HEADER_LEN)
            try:
                decoded_header = buf.decode("utf-8")
            except UnicodeDecodeError as e:
                raise DeserializationError("Invalid magic header. Decoder error: %s", e)

            if WHYLOGS_MAGIC_HEADER != decoded_header:
                raise DeserializationError(
                    f"Invalid magic header. Got: {decoded_header} but expecting: {WHYLOGS_MAGIC_HEADER}"
                )

            dataset_segment_header = read_delimited_protobuf(f, DatasetSegmentHeader)
            if dataset_segment_header.has_segments:
                logger.warning(
                    "File contains segments. Only first profile will be deserialized into this DatasetProfileView"
                )

            dataset_profile_header = read_delimited_protobuf(f, DatasetProfileHeader)
            if dataset_profile_header.ByteSize() == 0:
                raise DeserializationError("Missing valid dataset profile header")
            dataset_timestamp = datetime.fromtimestamp(dataset_profile_header.properties.dataset_timestamp / 1000.0)
            creation_timestamp = datetime.fromtimestamp(dataset_profile_header.properties.creation_timestamp / 1000.0)
            indexed_metric_paths = dataset_profile_header.indexed_metric_paths
            if len(indexed_metric_paths) < 1:
                logger.warning("Name index in the header is empty. Possible data corruption")

            start_offset = f.tell()

            columns = {}
            for col_name in sorted(dataset_profile_header.column_offsets.keys()):
                col_offsets = dataset_profile_header.column_offsets[col_name]
                all_metric_components: Dict[str, MetricComponentMessage] = {}
                for offset in col_offsets.offsets:
                    actual_offset = offset + start_offset
                    chunk_header = read_delimited_protobuf(f, proto_class_name=ChunkHeader, offset=actual_offset)
                    if chunk_header is None:
                        raise DeserializationError(
                            f"Missing or corrupt chunk header for column {col_name}. Offset: {actual_offset}"
                        )
                    if chunk_header.type != ChunkHeader.ChunkType.COLUMN:
                        raise DeserializationError(
                            f"Expecting chunk header type to be {ChunkHeader.ChunkType.COLUMN}, "
                            f"got {chunk_header.type}"
                        )

                    chunk_msg = ChunkMessage()
                    buf = f.read(chunk_header.length)
                    if len(buf) != chunk_header.length:
                        raise IOError(
                            f"Invalid message for {col_name}. Expecting buffer length of {chunk_header.length}, "
                            f"got {len(buf)}. "
                            f"Offset: {actual_offset}"
                        )
                    try:
                        chunk_msg.ParseFromString(buf)
                    except DecodeError:
                        raise DeserializationError(f"Failed to parse protobuf message for column: {col_name}")

                    for idx, metric_component in chunk_msg.metric_components.items():
                        full_name = indexed_metric_paths.get(idx)
                        if full_name is None:
                            raise ValueError(f"Missing metric name in the header. Index: {idx}")
                        all_metric_components[full_name] = metric_component

                column_msg = ColumnMessage(metric_components=all_metric_components)
                columns[col_name] = ColumnProfileView.from_protobuf(column_msg)
            return DatasetProfileView(
                columns=columns, dataset_timestamp=dataset_timestamp, creation_timestamp=creation_timestamp
            )

    def to_pandas(self, column_metric: Optional[str] = None, cfg: Optional[SummaryConfig] = None) -> pd.DataFrame:
        all_dicts = []
        for col_name, col in sorted(self._columns.items()):
            sum_dict = col.to_summary_dict(column_metric=column_metric, cfg=cfg)
            sum_dict["column"] = col_name
            sum_dict["type"] = SummaryType.COLUMN
            all_dicts.append(dict(sorted(sum_dict.items())))
        df = pd.DataFrame(all_dicts)
        return df.set_index("column")



import logging
import os
from pathlib import Path
from typing import Callable

logger = logging.getLogger(__name__)
_file_dir = Path(os.path.dirname(__file__))
_VIZ_DIR = _file_dir.parent.absolute()


def _get_template_path(html_file_name: str) -> str:
    template_path = os.path.abspath(os.path.join(_VIZ_DIR, "html", "templates", html_file_name))
    return template_path


def _get_compiled_template(template_name: str) -> "Callable":
    template_path = _get_template_path(template_name)
    try:
        from pybars import Compiler  # type: ignore
    except ImportError as e:
        logger.debug(e, exc_info=True)
        logger.warning("Unable to load pybars; install pybars3 to load profile directly from the current session ")
        return lambda _: None  # returns a no-op lambda

    with open(template_path, "r+t") as file_with_template:
        source = file_with_template.read()
    return Compiler().compile(source)


def get_compiled_template(template_name: str) -> "Callable":
    return _get_compiled_template(template_name=template_name)



from .column_profile import ColumnProfile
from .dataset_profile import DatasetProfile
from .datatypes import TypeMapper
from .metrics import MetricConfig
from .model_performance_metrics import ModelPerformanceMetrics
from .resolvers import Resolver
from .schema import ColumnSchema, DatasetSchema
from .segment import Segment
from .segmentation_partition import SegmentationPartition
from .view import WHYLOGS_MAGIC_HEADER, ColumnProfileView, DatasetProfileView

__ALL__ = [
    WHYLOGS_MAGIC_HEADER,
    # column
    ColumnProfile,
    DatasetProfile,
    DatasetSchema,
    ColumnSchema,
    # metric config
    MetricConfig,
    # model perf
    ModelPerformanceMetrics,
    # Typing
    TypeMapper,
    Resolver,
    # Views
    ColumnProfileView,
    DatasetProfileView,
    Segment,
    SegmentationPartition,
]



from abc import ABC, abstractmethod
from decimal import Decimal
from typing import Any, Generic, List, Optional, Type, TypeVar, Union

from whylogs.core.stubs import np

try:
    from pandas.core.api import CategoricalDtype
except:  # noqa
    CategoricalDtype = None  # type: ignore

NT = TypeVar("NT")


class DataType(ABC, Generic[NT]):
    def __init__(self, tpe: Type[NT]):
        self._tpe = tpe

    @property
    def return_type(self) -> Type[NT]:
        return self._tpe

    @classmethod
    def match(cls, dtype_or_type: Any) -> bool:
        if dtype_or_type is None:
            return False
        if hasattr(dtype_or_type, "type"):
            maybe_type = dtype_or_type.type  # type: ignore
        else:
            maybe_type = None

        return cls._do_match(dtype_or_type, maybe_type)

    @classmethod
    @abstractmethod
    def _do_match(cls, dtype_or_type: Any, maybe_type: Optional[Any]) -> bool:
        raise NotImplementedError


NumericalType = Union[int, float, np.number, Decimal]

NUMBER = TypeVar("NUMBER")


class Integral(DataType[int]):
    def __init__(self) -> None:
        super().__init__(int)

    @classmethod
    def _do_match(cls, dtype_or_type: Any, maybe_type: Optional[Any]) -> bool:
        if maybe_type:
            dtype_or_type = maybe_type  # type: ignore

        if not isinstance(dtype_or_type, type):
            return False

        if issubclass(dtype_or_type, (bool, int, np.number)):
            if np.issubdtype and np.issubdtype(dtype_or_type, np.floating):
                return False
            if issubclass(dtype_or_type, (np.datetime64, np.timedelta64)):
                return False
            else:
                return True
        return False


class Fractional(DataType[float]):
    def __init__(self) -> None:
        super().__init__(float)

    @classmethod
    def _do_match(cls, dtype_or_type: Any, maybe_type: Optional[Any]) -> bool:
        if maybe_type:
            dtype_or_type = maybe_type

        if not isinstance(dtype_or_type, type):
            return False

        return issubclass(dtype_or_type, (float, np.floating, Decimal))


class String(DataType[str]):
    def __init__(self) -> None:
        super().__init__(str)

    @classmethod
    def _do_match(cls, dtype_or_type: Any, maybe_type: Optional[Any]) -> bool:
        # Pandas Categorical is Strings
        if CategoricalDtype is not None and isinstance(dtype_or_type, CategoricalDtype):
            return True

        # handle pandas series. Anything 'object' is treated as a Series of strings
        if hasattr(dtype_or_type, "name"):
            if dtype_or_type.name == "object":  # type: ignore
                return True

        # use the underlying type within a dtype object
        if maybe_type:
            dtype_or_type = maybe_type

        if not isinstance(dtype_or_type, type):
            return False

        if issubclass(dtype_or_type, (str, np.unicode_)):
            return True

        return False


class AnyType(DataType[Any]):
    def __init__(self) -> None:
        super().__init__(type)

    def is_compatible(self, value: Any) -> bool:
        return True

    @classmethod
    def _do_match(cls, dtype_or_type: Any, maybe_type: Optional[Any]) -> bool:
        return True


class TypeMapper(ABC):
    @abstractmethod
    def __call__(self, dtype_or_type: Any) -> DataType:
        raise NotImplementedError


class StandardTypeMapper(TypeMapper):
    """Map a dtype (Pandas) or a Python type to a data type."""

    def __init__(self, custom_types: Optional[List[Type[DataType]]] = None):
        """

        Args:
            custom_types: List of additional DataType classes that you want to extend.
        """
        self._all_types = [Integral, Fractional, String]
        if custom_types is not None:
            self._all_types.extend(custom_types)

    def __call__(self, dtype_or_type: Any) -> DataType:
        for clz in self._all_types:
            if clz.match(dtype_or_type):  # type: ignore
                return clz()  # type: ignore
        return AnyType()



from typing import Any, Dict, List, Optional

from whylogs.api.logger.logger import Logger
from whylogs.core import DatasetProfile, DatasetSchema
from whylogs.core.stubs import pd


class TransientLogger(Logger):
    def __init__(self, schema: Optional[DatasetSchema] = None):
        super(TransientLogger, self).__init__(schema)

    def _get_matching_profiles(
        self, obj: Any = None, *, pandas: Optional[pd.DataFrame] = None, row: Optional[Dict[str, Any]] = None
    ) -> List[DatasetProfile]:
        return [DatasetProfile(schema=self._schema)]

    def flush(self) -> None:
        pass

    def close(self) -> None:
        pass



from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Dict, List, Optional

from whylogs.api.writer.writer import Writable
from whylogs.core import DatasetProfile, Segment
from whylogs.core.schema import DatasetSchema
from whylogs.core.stubs import pd


class ProfileStore(ABC):
    """
    A holder object for profiles.
    """

    @abstractmethod
    def get_matching_profile(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Dict[str, Any]] = None,
        segment: Optional[Segment] = None,
        schema: Optional[DatasetSchema] = None,
    ) -> List[DatasetProfile]:
        return [DatasetProfile(schema)]

    @abstractmethod
    def read(
        self,
        id: Optional[str],
        dataset_timestamp: Optional[datetime] = None,
        segment: Optional[Segment] = None,
        schema: Optional[DatasetSchema] = None,
    ) -> Optional[DatasetProfile]:
        return [DatasetProfile(schema)]

    @abstractmethod
    def write(
        self,
        data: Writable,
        model_id: Optional[str] = None,
        dataset_timestamp: Optional[datetime] = None,
        segment: Optional[Segment] = None,
        schema: Optional[DatasetSchema] = None,
    ) -> None:
        pass

    @abstractmethod
    def query(self, *args: Any, **kwargs: Any) -> List[DatasetProfile]:
        pass



from dataclasses import dataclass
from typing import Any, Dict

from whylogs.core.configs import SummaryConfig
from whylogs.core.metrics.metric_components import IntegralComponent
from whylogs.core.metrics.metrics import Metric, MetricConfig, OperationResult
from whylogs.core.preprocessing import PreprocessedColumn


@dataclass(frozen=True)
class TypeCountersMetric(Metric):
    integral: IntegralComponent
    fractional: IntegralComponent
    boolean: IntegralComponent
    string: IntegralComponent
    object: IntegralComponent

    @property
    def namespace(self) -> str:
        return "types"

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        return {
            "integral": self.integral.value,
            "fractional": self.fractional.value,
            "boolean": self.boolean.value,
            "string": self.string.value,
            "object": self.object.value,
        }

    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        if data.len <= 0:
            return OperationResult.ok(0)

        successes = 0

        integral_prev = self.integral.value
        integral = 0
        if data.numpy.ints is not None:
            integral += len(data.numpy.ints)
        elif data.list.ints is not None:
            integral += len(data.list.ints)
        successes += integral
        self.integral.set(integral + integral_prev)

        fractional = 0
        fractional_prev = self.fractional.value
        if data.numpy.floats is not None:
            fractional += len(data.numpy.floats)
        elif data.list.floats is not None:
            fractional += len(data.list.floats)
        successes += fractional
        self.fractional.set(fractional + fractional_prev)

        bool_count = self.boolean.value + data.bool_count
        successes += data.bool_count
        self.boolean.set(bool_count)

        string_count = 0
        string_count_prev = self.string.value
        if data.pandas.strings is not None:
            string_count += len(data.pandas.strings)
        elif data.list.strings is not None:
            string_count += len(data.list.strings)

        successes += string_count
        self.string.set(string_count + string_count_prev)

        objects = 0
        objects_prev = self.object.value
        if data.pandas.objs is not None:
            objects += len(data.pandas.objs)
        elif data.list.objs is not None:
            objects += len(data.list.objs)

        successes += objects
        self.object.set(objects + objects_prev)
        return OperationResult.ok(successes)

    @classmethod
    def zero(cls, config: MetricConfig) -> "TypeCountersMetric":
        return TypeCountersMetric(
            integral=IntegralComponent(0),
            fractional=IntegralComponent(0),
            boolean=IntegralComponent(0),
            string=IntegralComponent(0),
            object=IntegralComponent(0),
        )


@dataclass(frozen=True)
class ColumnCountsMetric(Metric):
    n: IntegralComponent
    null: IntegralComponent

    @property
    def namespace(self) -> str:
        return "counts"

    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        n: int = self.n.value
        null: int = self.null.value
        if data.len <= 0:
            return OperationResult.ok(0)
        n += data.len
        self.n.set(n)

        null += data.null_count
        self.null.set(null)
        return OperationResult.ok(data.len)

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        return {"n": self.n.value, "null": self.null.value}

    @classmethod
    def zero(cls, config: MetricConfig) -> "ColumnCountsMetric":
        return ColumnCountsMetric(n=IntegralComponent(0), null=IntegralComponent(0))
